# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Arch Schedule"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchSchedule
#  \ingroup ARCH
#  \brief The Schedule object and tools
#
#  This module provides tools to build Schedule objects.
#  Schedules are objects that can count and gather information
#  about objects in the document, and fill a spreadsheet with the result

import FreeCAD

from draftutils import params

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")
VERBOSE = True  # change this for silent recomputes


class _ArchScheduleDocObserver:
    "doc observer to monitor all recomputes"

    # https://forum.freecad.org/viewtopic.php?style=3&p=553377#p553377

    def __init__(self, doc, schedule):
        self.doc = doc
        self.schedule = schedule

    def slotRecomputedDocument(self, doc):
        if doc != self.doc:
            return
        try:
            self.schedule.Proxy.execute(self.schedule)
        except:
            pass


class _ArchSchedule:
    "the Arch Schedule object"

    def __init__(self, obj):

        self.setProperties(obj)
        obj.Proxy = self
        self.Type = "Schedule"

    def onDocumentRestored(self, obj):

        self.setProperties(obj)
        if hasattr(obj, "Result"):
            self.update_properties_0v21(obj)
        if hasattr(obj, "Description"):
            self.update_properties_1v1(obj)

    def update_properties_0v21(self, obj):
        from draftutils.messages import _log

        sp = obj.Result
        if sp is not None:
            self.setSchedulePropertySpreadsheet(sp, obj)
        obj.removeProperty("Result")
        _log("v0.21, " + obj.Name + ", removed property 'Result', and added property 'AutoUpdate'")
        if sp is not None:
            _log("v0.21, " + sp.Name + ", added property 'Schedule'")

    def update_properties_1v1(self, obj):
        from draftutils.messages import _log

        if obj.getTypeIdOfProperty("Description") == "App::PropertyStringList":
            obj.Operation = obj.Description
            obj.removeProperty("Description")
            _log("v1.1, " + obj.Name + ", renamed property 'Description' to 'Operation'")
        for prop in (
            "Operation",
            "Value",
            "Unit",
            "Objects",
            "Filter",
            "CreateSpreadsheet",
            "DetailedResults",
        ):
            obj.setGroupOfProperty(prop, "Schedule")

    def setProperties(self, obj):

        if not "Operation" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyStringList",
                "Operation",
                "Schedule",
                QT_TRANSLATE_NOOP("App::Property", "The operation column"),
                locked=True,
            )
        if not "Value" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyStringList",
                "Value",
                "Schedule",
                QT_TRANSLATE_NOOP("App::Property", "The values column"),
                locked=True,
            )
        if not "Unit" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyStringList",
                "Unit",
                "Schedule",
                QT_TRANSLATE_NOOP("App::Property", "The units column"),
                locked=True,
            )
        if not "Objects" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyStringList",
                "Objects",
                "Schedule",
                QT_TRANSLATE_NOOP("App::Property", "The objects column"),
                locked=True,
            )
        if not "Filter" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyStringList",
                "Filter",
                "Schedule",
                QT_TRANSLATE_NOOP("App::Property", "The filter column"),
                locked=True,
            )
        if not "CreateSpreadsheet" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyBool",
                "CreateSpreadsheet",
                "Schedule",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If True, a spreadsheet containing the results is recreated when needed",
                ),
                locked=True,
            )
        if not "DetailedResults" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyBool",
                "DetailedResults",
                "Schedule",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If True, additional lines with each individual object are added to the results",
                ),
                locked=True,
            )
        if not "AutoUpdate" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyBool",
                "AutoUpdate",
                "Schedule",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If True, the schedule and the associated spreadsheet are updated whenever the document is recomputed",
                ),
                locked=True,
            )
            obj.AutoUpdate = True

        # To add the doc observer:
        self.onChanged(obj, "AutoUpdate")

    def setSchedulePropertySpreadsheet(self, sp, obj):
        if not hasattr(sp, "Schedule"):
            sp.addProperty(
                "App::PropertyLink",
                "Schedule",
                "Arch",
                QT_TRANSLATE_NOOP("App::Property", "The BIM Schedule that uses this spreadsheet"),
                locked=True,
            )
        sp.Schedule = obj

    def getSpreadSheet(self, obj, force=False):
        """Get the spreadsheet and store it in self.spreadsheet.

        If force is True the spreadsheet is created if required.
        """
        try:  # Required as self.spreadsheet may get deleted.
            if (
                getattr(self, "spreadsheet", None) is not None
                and getattr(self.spreadsheet, "Schedule", None) == obj
            ):
                return self.spreadsheet
        except:
            pass
        else:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.TypeId == "Spreadsheet::Sheet" and getattr(o, "Schedule", None) == obj:
                    self.spreadsheet = o
                    return self.spreadsheet
            if force:
                self.spreadsheet = FreeCAD.ActiveDocument.addObject("Spreadsheet::Sheet", "Result")
                self.setSchedulePropertySpreadsheet(self.spreadsheet, obj)
                return self.spreadsheet
            else:
                return None

    def onChanged(self, obj, prop):

        if prop == "CreateSpreadsheet":
            if obj.CreateSpreadsheet:
                self.getSpreadSheet(obj, force=True)
            else:
                sp = self.getSpreadSheet(obj)
                if sp is not None:
                    FreeCAD.ActiveDocument.removeObject(sp.Name)
                    self.spreadsheet = None
        elif prop == "AutoUpdate":
            if obj.AutoUpdate:
                if getattr(self, "docObserver", None) is None:
                    self.docObserver = _ArchScheduleDocObserver(FreeCAD.ActiveDocument, obj)
                    FreeCAD.addDocumentObserver(self.docObserver)
            elif getattr(self, "docObserver", None) is not None:
                FreeCAD.removeDocumentObserver(self.docObserver)
                self.docObserver = None

    def setSpreadsheetData(self, obj, force=False):
        """Fills a spreadsheet with the stored data"""

        if not hasattr(self, "data"):
            self.execute(obj)
        if not hasattr(self, "data"):
            return
        if not self.data:
            return
        if not (obj.CreateSpreadsheet or force):
            return
        sp = self.getSpreadSheet(obj, force=True)
        widths = [sp.getColumnWidth(col) for col in ("A", "B", "C")]
        sp.clearAll()
        # clearAll resets the column widths:
        for col, width in zip(("A", "B", "C"), widths):
            sp.setColumnWidth(col, width)
        # set headers
        sp.set("A1", "Operation")
        sp.set("B1", "Value")
        sp.set("C1", "Unit")
        sp.setStyle("A1:C1", "bold", "add")
        # write contents
        for k, v in self.data.items():
            sp.set(k, v)
        # recompute
        sp.recompute()
        sp.purgeTouched()  # Remove the confusing blue checkmark from the spreadsheet.
        for o in sp.InList:  # Also recompute TechDraw views.
            o.TypeId == "TechDraw::DrawViewSpreadsheet"
            o.recompute()

    def execute(self, obj):

        # verify the data

        if not obj.Operation:
            # empty description column
            return
        for p in [obj.Value, obj.Unit, obj.Objects, obj.Filter]:
            # different number of items in each column
            if len(obj.Operation) != len(p):
                return

        self.data = {}  # store all results in self.data, so it lives even without spreadsheet
        self.li = 1  # row index - starts at 2 to leave 2 blank rows for the title

        for i in range(len(obj.Operation)):
            self.li += 1
            if not obj.Operation[i]:
                # blank line
                continue
            # write description
            self.data["A" + str(self.li)] = obj.Operation[i]
            if VERBOSE:
                l = "OPERATION: " + obj.Operation[i]
                print("")
                print(l)
                print(len(l) * "=")

            # build set of valid objects

            objs = obj.Objects[i]
            val = obj.Value[i]
            unit = obj.Unit[i]
            details = obj.DetailedResults
            ifcfile = None
            elts = None
            if val:
                import Draft
                import Arch

                if objs:
                    objs = objs.split(";")
                    objs = [FreeCAD.ActiveDocument.getObject(o) for o in objs]
                    objs = [o for o in objs if o is not None]
                else:
                    if hasattr(getattr(FreeCAD.ActiveDocument, "Proxy", None), "ifcfile"):
                        ifcfile = FreeCAD.ActiveDocument.Proxy.ifcfile
                    objs = FreeCAD.ActiveDocument.Objects
                if len(objs) == 1:
                    if hasattr(objs[0], "StepId"):
                        from nativeifc import ifc_tools

                        ifcfile = ifc_tools.get_ifcfile(objs[0])
                    # remove object itself if the object is a group
                    if objs[0].isDerivedFrom("App::DocumentObjectGroup"):
                        objs = objs[0].Group
                objs = Draft.get_group_contents(objs)
                objs = self.expandArrays(objs)
                # Remove included objects (e.g. walls that are part of another wall,
                # base geometry, etc)
                objs = Arch.pruneIncluded(objs, strict=True, silent=True)
                # Remove all schedules and spreadsheets:
                objs = [
                    o for o in objs if Draft.get_type(o) not in ["Schedule", "Spreadsheet::Sheet"]
                ]

                # filter elements

                if obj.Filter[i]:
                    if ifcfile:
                        elts = self.get_ifc_elements(ifcfile, obj.Filter[i])
                    else:
                        objs = self.apply_filter(objs, obj.Filter[i])

                # perform operation: count or retrieve property

                if ifcfile:
                    if elts:
                        self.update_from_elts(elts, val, unit, details)
                elif objs:
                    self.update_from_objs(objs, val, unit, details)

        self.setSpreadsheetData(obj)
        self.save_ifc_props(obj)

    def apply_filter(self, objs, filters):
        """Applies the given filters to the given list of objects"""

        nobjs = []
        for o in objs:
            props = [p.upper() for p in o.PropertiesList]
            ok = True
            for f in filters.split(";"):
                args = [a.strip() for a in f.strip().split(":")]
                if args[0][0] == "!":
                    inv = True
                    prop = args[0][1:].upper()
                else:
                    inv = False
                    prop = args[0].upper()
                fval = args[1].upper()
                if prop == "TYPE":
                    prop = "IFCTYPE"
                if inv:
                    if prop in props:
                        csprop = o.PropertiesList[props.index(prop)]
                        if fval in getattr(o, csprop).upper():
                            ok = False
                else:
                    if not (prop in props):
                        ok = False
                    else:
                        csprop = o.PropertiesList[props.index(prop)]
                        if not (fval in getattr(o, csprop).upper()):
                            ok = False
            if ok:
                nobjs.append(o)
        return nobjs

    def get_ifc_elements(self, ifcfile, filters):
        """Retrieves IFC elements corresponding to the given filters"""

        elts = []
        for el in ifcfile.by_type("IfcProduct"):
            ok = True
            for f in filters.split(";"):
                args = [a.strip() for a in f.strip().split(":")]
                if args[0][0] == "!":
                    inv = True
                    prop = args[0][1:]
                else:
                    inv = False
                    prop = args[0]
                fval = args[1]
                if prop.upper() in ["CLASS", "IFCCLASS", "IFCTYPE"]:
                    prop = "is_a"
                if inv:
                    if prop == "is_a":
                        if not fval.upper().startswith("IFC"):
                            fval = "Ifc" + fval
                        fval = fval.replace(" ", "")
                        if el.is_a(fval):
                            ok = False
                    else:
                        if prop in dir(el):
                            rval = getattr(el, prop)
                            if hasattr(rval, "id"):
                                if fval.startswith("#"):
                                    fval = int(fval[1:])
                            if rval == fval:
                                ok = False
                else:
                    if prop == "is_a":
                        if not fval.upper().startswith("IFC"):
                            fval = "Ifc" + fval
                        fval = fval.replace(" ", "")
                        if not el.is_a(fval):
                            ok = False
                    else:
                        if prop in dir(el):
                            rval = getattr(el, prop)
                            if hasattr(rval, "id"):
                                if fval.startswith("#"):
                                    fval = int(fval[1:])
                            if rval != fval:
                                ok = False
                        else:
                            ok = False
            if ok:
                elts.append(el)
        return elts

    def update_from_objs(self, objs, val, unit, details):
        """Updates the spreadsheet data from FreeCAD objects"""

        if val.upper() == "COUNT":
            val = len(objs)
            if VERBOSE:
                print(val, ",".join([o.Label for o in objs]))
            self.data["B" + str(self.li)] = str(val)
            if details:
                # additional blank line...
                self.li += 1
                self.data["A" + str(self.li)] = " "
        else:
            vals = val.split(".")
            if vals[0][0].islower():
                # old-style: first member is not a property
                vals = vals[1:]
            sumval = 0

            # get unit
            tp = None
            unit = None
            q = None
            if unit:
                unit = unit.replace("^", "")  # get rid of existing power symbol
                unit = unit.replace("2", "^2")
                unit = unit.replace("3", "^3")
                unit = unit.replace("²", "^2")
                unit = unit.replace("³", "^3")
                if "2" in unit:
                    tp = FreeCAD.Units.Area
                elif "3" in unit:
                    tp = FreeCAD.Units.Volume
                elif "deg" in unit:
                    tp = FreeCAD.Units.Angle
                else:
                    tp = FreeCAD.Units.Length

            # format value
            dv = params.get_param("Decimals", path="Units")
            fs = "{:." + str(dv) + "f}"  # format string
            for o in objs:
                if VERBOSE:
                    l = o.Name + " (" + o.Label + "):"
                    print(l + (40 - len(l)) * " ", end="")
                try:
                    d = o
                    for v in vals:
                        d = getattr(d, v)
                    if hasattr(d, "Value"):
                        d = d.Value
                except Exception:
                    t = translate("Arch", "Unable to retrieve value from object")
                    FreeCAD.Console.PrintWarning(t + ": " + o.Name + "." + ".".join(vals) + "\n")
                else:
                    if VERBOSE:
                        if tp and unit:
                            v = fs.format(FreeCAD.Units.Quantity(d, tp).getValueAs(unit).Value)
                            print(v, unit)
                        elif isinstance(d, str):
                            if d.replace(".", "", 1).isdigit():
                                print(fs.format(d))
                            else:
                                print(d)
                        else:
                            print(fs.format(d))
                    if details:
                        self.li += 1
                        self.data["A" + str(self.li)] = o.Name + " (" + o.Label + ")"
                        if tp and unit:
                            q = FreeCAD.Units.Quantity(d, tp)
                            self.data["B" + str(self.li)] = str(q.getValueAs(unit).Value)
                            self.data["C" + str(self.li)] = unit
                        else:
                            self.data["B" + str(self.li)] = str(d)

                    if sumval:
                        sumval += d
                    else:
                        sumval = d
            val = sumval
            if tp:
                q = FreeCAD.Units.Quantity(val, tp)

            # write data
            if details:
                self.li += 1
                self.data["A" + str(self.li)] = "TOTAL"
            if q and unit:
                self.data["B" + str(self.li)] = str(q.getValueAs(unit).Value)
                self.data["C" + str(self.li)] = unit
            else:
                self.data["B" + str(self.li)] = str(val)
            if VERBOSE:
                if tp and unit:
                    v = fs.format(FreeCAD.Units.Quantity(val, tp).getValueAs(unit).Value)
                    print("TOTAL:" + 34 * " " + v + " " + unit)
                elif isinstance(val, str):
                    if val.replace(".", "", 1).isdigit():
                        v = fs.format(val)
                        print("TOTAL:" + 34 * " " + v)
                    else:
                        print("TOTAL:" + 34 * " " + val)
                else:
                    v = fs.format(val)
                    print("TOTAL:" + 34 * " " + v)

    def update_from_elts(self, elts, val, unit, details):
        """Updates the spreadsheet data from IFC elements"""

        if val.upper() == "COUNT":
            val = len(elts)
            if VERBOSE:
                print("COUNT:", val, "(", ",".join(["#" + str(e.id()) for e in elts]), ")")
            self.data["B" + str(self.li)] = str(val)
            if details:
                # additional blank line...
                self.li += 1
                self.data["A" + str(self.li)] = " "
        else:
            total = 0
            for el in elts:
                if val in dir(el):
                    elval = getattr(el, val, "")
                    if isinstance(elval, tuple):
                        if len(elval) == 1:
                            elval = elval[0]
                        elif len(elval) == 0:
                            elval = ""
                    if hasattr(elval, "is_a") and elval.is_a("IfcRelationship"):
                        for att in dir(elval):
                            if att.startswith("Relating"):
                                targ = getattr(elval, att)
                                if targ != el:
                                    elval = targ
                                    break
                            elif att.startswith("Related"):
                                if not elval in getattr(elval, att):
                                    elval = str(getattr(elval, att))
                                    break
                    if details:
                        self.li += 1
                        name = el.Name if el.Name else ""
                        self.data["A" + str(self.li)] = "#" + str(el.id()) + name
                        self.data["B" + str(self.li)] = str(elval)
                        if VERBOSE:
                            print("#" + str(el.id()) + "." + val + " = " + str(elval))
                    if isinstance(elval, str) and elval.replace(".", "", 1).isdigit():
                        total += float(elval)
                    elif isinstance(elval, (int, float)):
                        total += elval
            if total:
                if details:
                    self.li += 1
                    self.data["A" + str(self.li)] = "TOTAL"
                self.data["B" + str(self.li)] = str(total)
                if VERBOSE:
                    print("TOTAL:", str(total))

    def create_ifc(self, obj, ifcfile, export=False):
        """Creates an IFC element for this object"""

        from nativeifc import ifc_tools  # lazy loading

        proj = ifcfile.by_type("IfcProject")[0]
        elt = ifc_tools.api_run("root.create_entity", ifcfile, ifc_class="IfcControl")
        ifc_tools.set_attribute(ifcfile, elt, "Name", obj.Label)
        ifc_tools.api_run(
            "project.assign_declaration", ifcfile, definitions=[elt], relating_context=proj
        )
        if not export:
            ifc_tools.add_properties(obj, ifcfile, elt)
        return elt

    def save_ifc_props(self, obj, ifcfile=None, elt=None):
        """Saves the object data to IFC"""

        from nativeifc import ifc_psets  # lazy loading

        ifc_psets.edit_pset(
            obj, "Operation", "::".join(obj.Operation), ifcfile=ifcfile, element=elt
        )
        ifc_psets.edit_pset(obj, "Value", "::".join(obj.Value), ifcfile=ifcfile, element=elt)
        ifc_psets.edit_pset(obj, "Unit", "::".join(obj.Unit), ifcfile=ifcfile, element=elt)
        ifc_psets.edit_pset(obj, "Objects", "::".join(obj.Objects), ifcfile=ifcfile, element=elt)
        ifc_psets.edit_pset(obj, "Filter", "::".join(obj.Filter), ifcfile=ifcfile, element=elt)

    def export_ifc(self, obj, ifcfile):
        """Exports the object to IFC (does not modify the FreeCAD object)."""

        elt = self.create_ifc(obj, ifcfile, export=True)
        self.save_ifc_props(obj, ifcfile, elt)
        return elt

    def dumps(self):

        return self.Type

    def loads(self, state):

        if state:
            self.Type = state

    def getIfcClass(self, obj):
        """gets the IFC class of this object"""

        if hasattr(obj, "IfcType"):
            return obj.IfcType
        elif hasattr(obj, "IfcRole"):
            return obj.IfcRole
        elif hasattr(obj, "IfcClass"):
            return obj.IfcClass
        else:
            return None

    def getArray(self, obj):
        "returns a count number if this object needs to be duplicated"

        import Draft

        elementCount = 0

        # The given object can belong to multiple arrays
        # o is a potential parent array of the given object
        for o in obj.InList:
            if Draft.getType(o) == "Array":
                elementCount += o.Count

        return elementCount

    def expandArrays(self, objs):
        """Expands array elements in the given list of objects"""

        expandedobjs = []

        for obj in objs:
            ifcClass = self.getIfcClass(obj)
            # This filters out the array object itself, which has no IFC class,
            # but leaves the array elements, which do have an IFC class.
            if ifcClass:
                expandedobjs.append(obj)
                # If the object is in an array, add it and the rest of its elements
                # to the list.
                array = self.getArray(obj)
                for i in range(1, array):  # The first element (0) was already added
                    expandedobjs.append(obj)

        return expandedobjs


class _ViewProviderArchSchedule:
    "A View Provider for Schedules"

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        if self.Object.AutoUpdate is False:
            import TechDrawGui

            return ":/icons/TechDraw_TreePageUnsync.svg"
        import Arch_rc

        return ":/icons/Arch_Schedule.svg"

    def isShow(self):
        return True

    def attach(self, vobj):
        self.Object = vobj.Object

    def setEdit(self, vobj, mode=0):
        if mode != 0:
            return None

        self.taskd = ArchScheduleTaskPanel(vobj.Object)
        if not self.taskd.form.isVisible():
            from PySide import QtCore

            QtCore.QTimer.singleShot(100, self.showEditor)
        return True

    def showEditor(self):

        if hasattr(self, "taskd"):
            self.taskd.form.show()

    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        return True

    def doubleClicked(self, vobj):
        self.edit()
        return True

    def setupContextMenu(self, vobj, menu):

        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return

        actionEdit = QtGui.QAction(translate("Arch", "Edit"), menu)
        QtCore.QObject.connect(actionEdit, QtCore.SIGNAL("triggered()"), self.edit)
        menu.addAction(actionEdit)

        if self.Object.CreateSpreadsheet is True:
            msg = translate("Arch", "Remove spreadsheet")
        else:
            msg = translate("Arch", "Attach spreadsheet")
        actionToggleSpreadsheet = QtGui.QAction(QtGui.QIcon(":/icons/Arch_Schedule.svg"), msg, menu)
        QtCore.QObject.connect(
            actionToggleSpreadsheet, QtCore.SIGNAL("triggered()"), self.toggleSpreadsheet
        )
        menu.addAction(actionToggleSpreadsheet)

    def edit(self):
        FreeCADGui.ActiveDocument.setEdit(self.Object, 0)

    def toggleSpreadsheet(self):
        self.Object.CreateSpreadsheet = not self.Object.CreateSpreadsheet

    def claimChildren(self):
        if hasattr(self, "Object"):
            return [self.Object.Proxy.getSpreadSheet(self.Object)]

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def getDisplayModes(self, vobj):
        return ["Default"]

    def getDefaultDisplayMode(self):
        return "Default"

    def setDisplayMode(self, mode):
        return mode


class ArchScheduleTaskPanel:
    """The editmode TaskPanel for Schedules"""

    def __init__(self, obj=None):
        """Sets the panel up"""

        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/ArchSchedule.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Arch_Schedule.svg"))

        # set icons
        self.form.buttonAdd.setIcon(QtGui.QIcon(":/icons/list-add.svg"))
        self.form.buttonDel.setIcon(QtGui.QIcon(":/icons/list-remove.svg"))
        self.form.buttonClear.setIcon(QtGui.QIcon(":/icons/delete.svg"))
        self.form.buttonImport.setIcon(QtGui.QIcon(":/icons/document-open.svg"))
        self.form.buttonExport.setIcon(QtGui.QIcon(":/icons/document-save.svg"))
        self.form.buttonSelect.setIcon(QtGui.QIcon(":/icons/edit-select-all.svg"))

        # restore widths
        self.form.list.setColumnWidth(0, params.get_param_arch("ScheduleColumnWidth0"))
        self.form.list.setColumnWidth(1, params.get_param_arch("ScheduleColumnWidth1"))
        self.form.list.setColumnWidth(2, params.get_param_arch("ScheduleColumnWidth2"))
        self.form.list.setColumnWidth(3, params.get_param_arch("ScheduleColumnWidth3"))
        w = params.get_param_arch("ScheduleDialogWidth")
        h = params.get_param_arch("ScheduleDialogHeight")
        self.form.resize(w, h)

        # restore default states
        self.form.checkAutoUpdate.setChecked(PARAMS.GetBool("ScheduleAutoUpdate", False))

        # set delegate - Not using custom delegates for now...
        # self.form.list.setItemDelegate(ScheduleDelegate())
        # self.form.list.setEditTriggers(QtGui.QAbstractItemView.DoubleClicked)

        # connect slots
        self.form.buttonAdd.clicked.connect(self.add)
        self.form.buttonDel.clicked.connect(self.remove)
        self.form.buttonClear.clicked.connect(self.clear)
        self.form.buttonImport.clicked.connect(self.importCSV)
        self.form.buttonExport.clicked.connect(self.export)
        self.form.buttonSelect.clicked.connect(self.select)
        self.form.buttonBox.accepted.connect(self.accept)
        self.form.buttonBox.rejected.connect(self.reject)
        self.form.rejected.connect(self.reject)
        self.form.list.clearContents()

        if self.obj:
            # for p in [obj.Value,obj.Unit,obj.Objects,obj.Filter]:
            #    if len(obj.Operation) != len(p):
            #        return
            self.form.list.setRowCount(len(obj.Operation))
            for i in range(5):
                for j in range(len(obj.Operation)):
                    try:
                        text = [obj.Operation, obj.Value, obj.Unit, obj.Objects, obj.Filter][i][j]
                    except:
                        text = ""
                    item = QtGui.QTableWidgetItem(text)
                    self.form.list.setItem(j, i, item)
            self.form.lineEditName.setText(self.obj.Label)
            self.form.checkSpreadsheet.setChecked(self.obj.CreateSpreadsheet)
            self.form.checkDetailed.setChecked(self.obj.DetailedResults)
            self.form.checkAutoUpdate.setChecked(self.obj.AutoUpdate)

        # center over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        self.form.show()

    def add(self):
        """Adds a new row below the last one"""

        self.form.list.insertRow(self.form.list.currentRow() + 1)

    def remove(self):
        """Removes the current row"""

        if self.form.list.currentRow() >= 0:
            self.form.list.removeRow(self.form.list.currentRow())

    def clear(self):
        """Clears the list"""

        self.form.list.clearContents()
        self.form.list.setRowCount(0)

    def importCSV(self):
        """Imports a CSV file"""

        filename = QtGui.QFileDialog.getOpenFileName(
            QtGui.QApplication.activeWindow(),
            translate("Arch", "Import CSV file"),
            None,
            "CSV files (*.csv *.CSV)",
        )
        if filename:
            filename = filename[0]
            self.form.list.clearContents()
            import csv

            with open(filename, "r") as csvfile:
                r = 0
                for row in csv.reader(csvfile):
                    self.form.list.insertRow(r)
                    for i in range(5):
                        if len(row) > i:
                            t = row[i]
                            # t = t.replace("²","^2")
                            # t = t.replace("³","^3")
                            self.form.list.setItem(r, i, QtGui.QTableWidgetItem(t))
                    r += 1

    def export(self):
        """Exports the results as MD or CSV"""

        # commit latest changes
        self.writeValues()

        # tests
        if not ("Up-to-date" in self.obj.State):
            self.obj.Proxy.execute(self.obj)
        if not hasattr(self.obj.Proxy, "data"):
            return
        if not self.obj.Proxy.data:
            return

        filename = QtGui.QFileDialog.getSaveFileName(
            QtGui.QApplication.activeWindow(),
            translate("Arch", "Export CSV file"),
            None,
            "Comma-separated values (*.csv);;TAB-separated values (*.tsv);;Markdown (*.md)",
        )
        if filename:
            filt = filename[1]
            filename = filename[0]
            # add missing extension
            if (
                (not filename.lower().endswith(".csv"))
                and (not filename.lower().endswith(".tsv"))
                and (not filename.lower().endswith(".md"))
            ):
                if "csv" in filt:
                    filename += ".csv"
                elif "tsv" in filt:
                    filename += ".tsv"
                else:
                    filename += ".md"
            if filename.lower().endswith(".csv"):
                self.exportCSV(filename, delimiter=",")
            elif filename.lower().endswith(".tsv"):
                self.exportCSV(filename, delimiter="\t")
            elif filename.lower().endswith(".md"):
                self.exportMD(filename)
            else:
                FreeCAD.Console.PrintError(
                    translate("Arch", "Unable to recognize that file type") + ":" + filename + "\n"
                )

    def getRows(self):
        """get the rows that contain data"""

        rows = []
        if hasattr(self.obj.Proxy, "data") and self.obj.Proxy.data:
            for key in self.obj.Proxy.data.keys():
                n = key[1:]
                if not n in rows:
                    rows.append(n)
        rows.sort(key=int)
        return rows

    def exportCSV(self, filename, delimiter="\t"):
        """Exports the results as a CSV/TSV file"""

        import csv

        with open(filename, "w") as csvfile:
            csvfile = csv.writer(csvfile, delimiter=delimiter)
            csvfile.writerow(
                [
                    translate("Arch", "Operation"),
                    translate("Arch", "Value"),
                    translate("Arch", "Unit"),
                ]
            )
            if self.obj.DetailedResults:
                csvfile.writerow(["", "", ""])
            for i in self.getRows():
                r = []
                for j in ["A", "B", "C"]:
                    if j + i in self.obj.Proxy.data:
                        r.append(str(self.obj.Proxy.data[j + i]))
                    else:
                        r.append("")
                csvfile.writerow(r)
        print("successfully exported ", filename)

    def exportMD(self, filename):
        """Exports the results as a Markdown file"""

        with open(filename, "w") as mdfile:
            mdfile.write(
                "| "
                + translate("Arch", "Operation")
                + " | "
                + translate("Arch", "Value")
                + " | "
                + translate("Arch", "Unit")
                + " |\n"
            )
            mdfile.write("| --- | --- | --- |\n")
            if self.obj.DetailedResults:
                mdfile.write("| | | |\n")
            for i in self.getRows():
                r = []
                for j in ["A", "B", "C"]:
                    if j + i in self.obj.Proxy.data:
                        r.append(str(self.obj.Proxy.data[j + i]))
                    else:
                        r.append("")
                mdfile.write("| " + " | ".join(r) + " |\n")
        print("successfully exported ", filename)

    def select(self):
        """Adds selected objects to current row"""

        if self.form.list.currentRow() >= 0:
            sel = ""
            for o in FreeCADGui.Selection.getSelection():
                if o != self.obj:
                    if sel:
                        sel += ";"
                    sel += o.Name
            if sel:
                self.form.list.setItem(self.form.list.currentRow(), 3, QtGui.QTableWidgetItem(sel))

    def accept(self):
        """Saves the changes and closes the dialog"""

        # store widths
        params.set_param_arch("ScheduleColumnWidth0", self.form.list.columnWidth(0))
        params.set_param_arch("ScheduleColumnWidth1", self.form.list.columnWidth(1))
        params.set_param_arch("ScheduleColumnWidth2", self.form.list.columnWidth(2))
        params.set_param_arch("ScheduleColumnWidth3", self.form.list.columnWidth(3))
        params.set_param_arch("ScheduleDialogWidth", self.form.width())
        params.set_param_arch("ScheduleDialogHeight", self.form.height())

        # store default states
        PARAMS.SetBool("ScheduleAutoUpdate", self.form.checkAutoUpdate.isChecked())

        # commit values
        self.writeValues()
        self.form.hide()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def reject(self):
        """Close dialog without saving"""

        self.form.hide()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def writeValues(self):
        """commits values and recalculate"""

        if not self.obj:
            import Arch

            self.obj = Arch.makeSchedule()
        lists = [[], [], [], [], []]
        for i in range(self.form.list.rowCount()):
            for j in range(5):
                cell = self.form.list.item(i, j)
                if cell:
                    lists[j].append(cell.text())
                else:
                    lists[j].append("")
        FreeCAD.ActiveDocument.openTransaction("Edited Schedule")
        self.obj.Operation = lists[0]
        self.obj.Value = lists[1]
        self.obj.Unit = lists[2]
        self.obj.Objects = lists[3]
        self.obj.Filter = lists[4]
        self.obj.Label = self.form.lineEditName.text()
        self.obj.DetailedResults = self.form.checkDetailed.isChecked()
        self.obj.CreateSpreadsheet = self.form.checkSpreadsheet.isChecked()
        self.obj.AutoUpdate = self.form.checkAutoUpdate.isChecked()
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Jakub Michalski <jakub.j.michalski[at]gmail.com>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix amplitude"
__author__ = "Jakub Michalski"
__url__ = "https://www.freecad.org"


def get_analysis_types():
    return ["static"]


def get_sets_name():
    return "constraints_initial_temperature_node_sets"


def get_constraint_title():
    return "Final temperature constraint"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_constraint(f, femobj, inittemp_obj, ccxwriter):
    if inittemp_obj.EnableFinalTemperature:
        # floats read from ccx should use {:.13G}, see comment in writer module

        finaltemp = inittemp_obj.FinalTemperature.getValueAs("K")

        if inittemp_obj.EnableAmplitude:
            f.write(f"*TEMPERATURE, AMPLITUDE={inittemp_obj.Name}\n")
        else:
            f.write("*TEMPERATURE\n")
        if inittemp_obj.References:
            f.write(f"{inittemp_obj.Name},{finaltemp}\n")
        else:
            f.write(f"{ccxwriter.ccx_nall},{finaltemp}\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to convert Draft objects to Sketches and back.

Many Draft objects will be converted to a single non-constrained Sketch.

However, a single sketch with disconnected traces will be converted
into several individual Draft objects.
"""
## @package gui_draft2sketch
# \ingroup draftguitools
# \brief Provides GUI tools to convert Draft objects to Sketches and back.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
from draftutils.messages import _msg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Draft2Sketch(gui_base_original.Modifier):
    """Gui Command for the Draft2Sketch tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Draft2Sketch",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Draft2Sketch", "Draft to Sketch"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Draft2Sketch",
                "Converts bidirectionally between Draft objects and sketches.\nMultiple selected Draft objects are converted into a single sketch.\nHowever, a single sketch with disconnected traces is converted into several individual Draft objects.",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Convert Draft/Sketch")
        if not self.ui:
            return
        if not Gui.Selection.getSelection():
            self.ui.selectUi(on_close_call=self.finish)
            _msg(translate("draft", "Select an object to convert"))
            self.call = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)
        else:
            self.proceed()

    def proceed(self):
        """Proceed with the command if one object was selected."""
        if self.call is not None:
            self.end_callbacks(self.call)
        sel = Gui.Selection.getSelection()
        allSketches = True
        allDraft = True
        Gui.addModule("Draft")
        for obj in sel:
            if obj.isDerivedFrom("Sketcher::SketchObject"):
                allDraft = False
            elif obj.isDerivedFrom("Part::Part2DObjectPython") or obj.isDerivedFrom(
                "Part::Feature"
            ):
                allSketches = False
            else:
                allDraft = False
                allSketches = False

        if not sel:
            return
        elif allDraft:
            _cmd = "Draft.make_sketch"
            _cmd += "("
            _cmd += "FreeCADGui.Selection.getSelection(), "
            _cmd += "autoconstraints=True"
            _cmd += ")"
            _cmd_list = ["sk = " + _cmd, "FreeCAD.ActiveDocument.recompute()"]
            self.commit(translate("draft", "Convert to Sketch"), _cmd_list)
        elif allSketches:
            n = 0
            _cmd_list = list()
            for o in sel:
                _cmd = "Draft.draftify"
                _cmd += "("
                _cmd += "FreeCAD.ActiveDocument." + o.Name + ", "
                _cmd += "delete=False"
                _cmd += ")"
                _cmd_list.append("df" + str(n) + " = " + _cmd)
                n += 1

            _cmd_list.append("FreeCAD.ActiveDocument.recompute()")
            self.commit(translate("draft", "Convert to Draft"), _cmd_list)
        else:
            _cmd_list = list()
            n = 0
            for obj in sel:
                _cmd_df = "Draft.draftify"
                _cmd_df += "("
                _cmd_df += "FreeCAD.ActiveDocument." + obj.Name + ", "
                _cmd_df += "delete=False"
                _cmd_df += ")"

                _cmd_sk = "Draft.make_sketch"
                _cmd_sk += "("
                _cmd_sk += "FreeCAD.ActiveDocument." + obj.Name + ", "
                _cmd_sk += "autoconstraints=True"
                _cmd_sk += ")"

                if obj.isDerivedFrom("Sketcher::SketchObject"):
                    _cmd_list.append("obj" + str(n) + " = " + _cmd_df)
                elif obj.isDerivedFrom("Part::Part2DObjectPython") or obj.isDerivedFrom(
                    "Part::Feature"
                ):
                    _cmd_list.append("obj" + str(n) + " = " + _cmd_sk)
                # elif obj.isDerivedFrom("Part::Feature"):
                #    # if (len(obj.Shape.Wires) == 1
                #    #     or len(obj.Shape.Edges) == 1):
                #    _cmd_list.append("obj" + str(n) + " = " + _cmd_sk)
                n += 1
            _cmd_list.append("FreeCAD.ActiveDocument.recompute()")
            self.commit(translate("draft", "Convert Draft/Sketch"), _cmd_list)
        self.finish()


Gui.addCommand("Draft_Draft2Sketch", Draft2Sketch())

## @}



# ***************************************************************************
# *   Copyright (c) 2023 edi <edi271@a1.net>                                *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw HoleShaftFit GuiCommand."""

__title__ = "TechDrawTools.CommandHoleShaftFit"
__author__ = "edi"
__url__ = "https://www.freecad.org"
__version__ = "00.01"
__date__ = "2023/02/07"

from PySide.QtCore import QT_TRANSLATE_NOOP
from PySide import QtGui
from PySide.QtGui import QMessageBox

import FreeCAD as App
import FreeCADGui as Gui

import TechDrawTools

translate = App.Qt.translate


class CommandHoleShaftFit:
    """Adds a hole or shaft fit to a selected dimension."""

    def GetResources(self):
        """Return a dictionary with data that will be used by the button or menu item."""
        return {
            "Pixmap": "actions/TechDraw_HoleShaftFit.svg",
            "Accel": "",
            "MenuText": QT_TRANSLATE_NOOP(
                "TechDraw_HoleShaftFit", "Hole/Shaft Fit"
            ),
            "ToolTip": QT_TRANSLATE_NOOP(
                "TechDraw_HoleShaftFit",
                "Adds a hole or shaft fit to a selected length or diameter dimension"
            ),
        }

    def Activated(self):
        """Run the following code when the command is activated (button press)."""
        sel = Gui.Selection.getSelectionEx()
        # if sel and sel[0].Object.TypeId == 'TechDraw::DrawViewDimension':
        if sel[0].Object.TypeId == "TechDraw::DrawViewDimension":
            self.ui = TechDrawTools.TaskHoleShaftFit(sel)
            Gui.Control.showDialog(self.ui)
        else:
            msgBox = QtGui.QMessageBox()
            msgTitle = translate(
                "TechDraw_HoleShaftFit", "Add a hole or shaft fit to a dimension"
            )
            msg = translate(
                "TechDraw_HoleShaftFit",
                "Select one length dimension or diameter dimension and retry",
            )
            msgBox.setText(msg)
            msgBox.setWindowTitle(msgTitle)
            msgBox.exec_()

    def IsActive(self):
        """Return True when the command should be active or False when it should be disabled (greyed)."""
        if App.ActiveDocument:
            sel = Gui.Selection.getSelectionEx()
            return (
                TechDrawTools.TDToolsUtil.havePage()
                and TechDrawTools.TDToolsUtil.haveView()
                and len(sel) == 1
            )
        else:
            return False

#
# The command must be "registered" with a unique name by calling its class.
Gui.addCommand('TechDraw_HoleShaftFit', CommandHoleShaftFit())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import time
import hashlib
import logging
from collections import OrderedDict
from typing import Any, Dict, Set, NamedTuple, Optional, Tuple

# For type hinting Asset and AssetUri to avoid circular imports
# from typing import TYPE_CHECKING
# if TYPE_CHECKING:
#     from .asset import Asset
#     from .uri import AssetUri

logger = logging.getLogger(__name__)


class CacheKey(NamedTuple):
    store_name: str
    asset_uri_str: str
    raw_data_hash: int
    dependency_signature: Tuple


class CachedAssetEntry(NamedTuple):
    asset: Any  # Actually type Asset
    size_bytes: int  # Estimated size of the raw_data
    timestamp: float  # For LRU, or just use OrderedDict nature


class AssetCache:
    def __init__(self, max_size_bytes: int = 100 * 1024 * 1024):  # Default 100MB
        self.max_size_bytes: int = max_size_bytes
        self.current_size_bytes: int = 0

        self._cache: Dict[CacheKey, CachedAssetEntry] = {}
        self._lru_order: OrderedDict[CacheKey, None] = OrderedDict()

        self._cache_dependents_map: Dict[str, Set[CacheKey]] = {}
        self._cache_dependencies_map: Dict[CacheKey, Set[str]] = {}

    def _evict_lru(self):
        while self.current_size_bytes > self.max_size_bytes and self._lru_order:
            oldest_key, _ = self._lru_order.popitem(last=False)
            if oldest_key in self._cache:
                evicted_entry = self._cache.pop(oldest_key)
                self.current_size_bytes -= evicted_entry.size_bytes
                logger.debug(
                    f"Cache Evict (LRU): {oldest_key}, "
                    f"size {evicted_entry.size_bytes}. "
                    f"New size: {self.current_size_bytes}"
                )
                self._remove_key_from_dependency_maps(oldest_key)

    def _remove_key_from_dependency_maps(self, cache_key_to_remove: CacheKey):
        direct_deps_of_removed = self._cache_dependencies_map.pop(cache_key_to_remove, set())
        for dep_uri_str in direct_deps_of_removed:
            if dep_uri_str in self._cache_dependents_map:
                self._cache_dependents_map[dep_uri_str].discard(cache_key_to_remove)
                if not self._cache_dependents_map[dep_uri_str]:
                    del self._cache_dependents_map[dep_uri_str]

    def get(self, key: CacheKey) -> Optional[Any]:
        if key in self._cache:
            self._lru_order.move_to_end(key)
            logger.debug(f"Cache HIT: {key}")
            return self._cache[key].asset
        logger.debug(f"Cache MISS: {key}")
        return None

    def put(
        self,
        key: CacheKey,
        asset: Any,
        raw_data_size_bytes: int,
        direct_dependency_uri_strs: Set[str],
    ):
        if key in self._cache:
            self._remove_key_from_dependency_maps(key)
            self.current_size_bytes -= self._cache[key].size_bytes
            del self._cache[key]
            self._lru_order.pop(key, None)

        if raw_data_size_bytes > self.max_size_bytes:
            logger.warning(
                f"Asset {key.asset_uri_str} (size {raw_data_size_bytes}) "
                f"too large for cache (max {self.max_size_bytes}). Not caching."
            )
            return

        self.current_size_bytes += raw_data_size_bytes
        entry = CachedAssetEntry(asset=asset, size_bytes=raw_data_size_bytes, timestamp=time.time())
        self._cache[key] = entry
        self._lru_order[key] = None
        self._lru_order.move_to_end(key)

        self._cache_dependencies_map[key] = direct_dependency_uri_strs
        for dep_uri_str in direct_dependency_uri_strs:
            self._cache_dependents_map.setdefault(dep_uri_str, set()).add(key)

        logger.debug(
            f"Cache PUT: {key}, size {raw_data_size_bytes}. "
            f"Total cache size: {self.current_size_bytes}"
        )
        self._evict_lru()

    def invalidate_for_uri(self, updated_asset_uri_str: str):
        keys_to_remove_from_cache: Set[CacheKey] = set()
        invalidation_queue: list[str] = [updated_asset_uri_str]
        processed_uris_for_invalidation_round: Set[str] = set()

        while invalidation_queue:
            current_uri_to_check_str = invalidation_queue.pop(0)
            if current_uri_to_check_str in processed_uris_for_invalidation_round:
                continue
            processed_uris_for_invalidation_round.add(current_uri_to_check_str)

            for ck in list(self._cache.keys()):
                if ck.asset_uri_str == current_uri_to_check_str:
                    keys_to_remove_from_cache.add(ck)

            dependent_cache_keys = self._cache_dependents_map.get(
                current_uri_to_check_str, set()
            ).copy()

            for dep_ck in dependent_cache_keys:
                if dep_ck not in keys_to_remove_from_cache:
                    keys_to_remove_from_cache.add(dep_ck)
                    parent_uri_of_dep_ck = dep_ck.asset_uri_str
                    if parent_uri_of_dep_ck not in processed_uris_for_invalidation_round:
                        invalidation_queue.append(parent_uri_of_dep_ck)

        for ck_to_remove in keys_to_remove_from_cache:
            if ck_to_remove in self._cache:
                entry_to_remove = self._cache.pop(ck_to_remove)
                self.current_size_bytes -= entry_to_remove.size_bytes
                self._lru_order.pop(ck_to_remove, None)
                self._remove_key_from_dependency_maps(ck_to_remove)

        if keys_to_remove_from_cache:
            logger.debug(
                f"Cache invalidated for URI '{updated_asset_uri_str}' and "
                f"its dependents. Removed {len(keys_to_remove_from_cache)} "
                f"entries. New size: {self.current_size_bytes}"
            )

    def clear(self):
        self._cache.clear()
        self._lru_order.clear()
        self._cache_dependents_map.clear()
        self._cache_dependencies_map.clear()
        self.current_size_bytes = 0
        logger.info("AssetCache cleared.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "CompoundTools._CommandCompoundFilter"
__author__ = "DeepSOIC, Bernd Hahnebach"
__url__ = "https://www.freecad.org"
__doc__ = "Compound Filter: remove some children from a compound (features)."


import FreeCAD

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui
    from PySide import QtCore

    # translation-related code
    # (see forum thread "A new Part tool is being born... JoinFeatures!"
    # https://forum.freecad.org/viewtopic.php?f=22&t=11112&start=30#p90239 )
    try:
        _fromUtf8 = QtCore.QString.fromUtf8
    except Exception:

        def _fromUtf8(s):
            return s

    translate = FreeCAD.Qt.translate


# command class
class _CommandCompoundFilter:
    "Command to create CompoundFilter feature"

    def GetResources(self):
        return {
            "Pixmap": "Part_CompoundFilter",
            "MenuText": QtCore.QT_TRANSLATE_NOOP("Part_CompoundFilter", "Compound Filter"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Part_CompoundFilter",
                "Filters out objects from the selected compound "
                "by characteristics like volume,\n"
                "area, or length, or by choosing specific items.\n"
                "If a second object is selected, it will be used "
                "as reference, for example,\n"
                "for collision or distance filtering.",
            ),
        }

    def Activated(self):
        if (
            len(FreeCADGui.Selection.getSelection()) == 1
            or len(FreeCADGui.Selection.getSelection()) == 2
        ):
            cmdCreateCompoundFilter(name="CompoundFilter")
        else:
            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(
                translate(
                    "Part_CompoundFilter",
                    "First select a shape that is a compound. "
                    "If a second object is selected (optional) "
                    "it will be treated as a stencil.",
                    None,
                )
            )
            mb.setWindowTitle(translate("Part_CompoundFilter", "Bad Selection", None))
            mb.exec_()

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("Part_CompoundFilter", _CommandCompoundFilter())


# helper
def cmdCreateCompoundFilter(name):
    sel = FreeCADGui.Selection.getSelection()
    FreeCAD.ActiveDocument.openTransaction("Create CompoundFilter")
    FreeCADGui.addModule("CompoundTools.CompoundFilter")
    FreeCADGui.doCommand(
        "f = CompoundTools.CompoundFilter.makeCompoundFilter(name = '" + name + "')"
    )
    FreeCADGui.doCommand("f.Base = App.ActiveDocument." + sel[0].Name)
    if len(sel) == 2:
        FreeCADGui.doCommand("f.Stencil = App.ActiveDocument." + sel[1].Name)
        FreeCADGui.doCommand("f.Stencil.ViewObject.hide()")
        FreeCADGui.doCommand("f.FilterType = 'collision-pass'")
    else:
        FreeCADGui.doCommand("f.FilterType = 'window-volume'")

    try:
        FreeCADGui.doCommand("f.Proxy.execute(f)")
        FreeCADGui.doCommand("f.purgeTouched()")
    except Exception as err:
        if hasattr(err, "message"):
            error_string = err.message
        else:
            error_string = err
        mb = QtGui.QMessageBox()
        mb.setIcon(mb.Icon.Warning)
        error_text1 = translate("Part_CompoundFilter", "Computing the result failed with an error:")
        error_text2 = translate(
            "Part_CompoundFilter",
            "Click 'Continue' to create the feature anyway, or 'Abort' to cancel.",
        )
        mb.setText(error_text1 + "\n\n" + str(err) + "\n\n" + error_text2)
        mb.setWindowTitle(translate("Part_CompoundFilter", "Bad Selection", None))
        btnAbort = mb.addButton(QtGui.QMessageBox.StandardButton.Abort)
        btnOK = mb.addButton(
            translate("Part_SplitFeatures", "Continue", None),
            QtGui.QMessageBox.ButtonRole.ActionRole,
        )
        mb.setDefaultButton(btnOK)

        mb.exec_()

        if mb.clickedButton() is btnAbort:
            FreeCAD.ActiveDocument.abortTransaction()
            return

    FreeCADGui.doCommand("f.Base.ViewObject.hide()")
    FreeCADGui.doCommand("f = None")

    FreeCAD.ActiveDocument.commitTransaction()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import FreeCAD


class RotaryToolBitMixin:
    """
    Mixin class for rotary tool bits.
    Provides methods for accessing diameter and length from the shape.
    """

    def can_rotate(self) -> bool:
        return True

    def get_diameter(self) -> FreeCAD.Units.Quantity:
        """
        Get the diameter of the rotary tool bit from the shape.
        """
        return self.obj.Diameter

    def set_diameter(self, diameter: FreeCAD.Units.Quantity):
        """
        Set the diameter of the rotary tool bit on the shape.
        """
        if not isinstance(diameter, FreeCAD.Units.Quantity):
            raise ValueError("Diameter must be a FreeCAD Units.Quantity")
        self.obj.Diameter = diameter

    def get_length(self) -> FreeCAD.Units.Quantity:
        """
        Get the length of the rotary tool bit from the shape.
        """
        return self.obj.Length

    def set_length(self, length: FreeCAD.Units.Quantity):
        """
        Set the length of the rotary tool bit on the shape.
        """
        if not isinstance(length, FreeCAD.Units.Quantity):
            raise ValueError("Length must be a FreeCAD Units.Quantity")
        self.obj.Length = length



#!/usr/bin/env python
# SPDX-License-Identifier: LGPL-2.1-or-later

# basic test script for PropertyTopoShapeList

import FreeCAD as App
import Part
import os
import tempfile
import unittest


class TopoShapeListTest(unittest.TestCase):
    def setUp(self):
        """Creates a document and a FeaturePython that has a PropertyTopoShapeList property"""
        print("TopoShapeListTest - setting up")
        TempPath = tempfile.gettempdir()
        self.fileName = TempPath + os.sep + "TopoShapeListTest.FCStd"
        self.objName = "TestObject"
        App.newDocument("TopoShapeList")
        App.setActiveDocument("TopoShapeList")
        doc = App.newDocument()
        obj = doc.addObject("App::FeaturePython", self.objName)
        obj.addProperty("Part::PropertyTopoShapeList", "Shapes")
        box = Part.makeBox(1, 1, 1)
        box2 = Part.makeBox(1, 1, 2)
        box3 = Part.makeBox(1, 1, 3)
        obj.Shapes = [box, box2, box3]
        doc.saveAs(self.fileName)
        App.closeDocument(doc.Name)
        del obj, doc, box, box2, box3
        print("TopoShapeListTest: setUp complete")

    def tearDown(self):
        print("TopoShapeListTest finished")
        App.closeDocument("TopoShapeList")

    def testMakeTopoShapeList(self):
        """Tests PropertyTopoShapeList"""
        print("running TopoShapeListTest")
        doc = App.openDocument(self.fileName)
        doc.UndoMode = 1
        obj = doc.getObject(self.objName)
        boxes = obj.Shapes

        maxError = 0.0000001
        error = abs(1.0 - boxes[0].Volume)
        self.assertLessEqual(
            error, maxError, "TopoShapeList entry 0 has wrong volume: {0}".format(boxes[0].Volume)
        )
        error = abs(2.0 - boxes[1].Volume)
        self.assertLessEqual(
            error, maxError, "TopoShapeList entry 1 has wrong volume: {0}".format(boxes[1].Volume)
        )
        error = abs(3.0 - boxes[2].Volume)
        self.assertLessEqual(
            error, maxError, "TopoShapeList entry 2 has wrong volume: {0}".format(boxes[2].Volume)
        )

        twoboxes = [boxes[1], boxes[2]]
        doc.openTransaction("Change shapes")
        obj.Shapes = twoboxes
        doc.commitTransaction()
        self.assertEqual(
            len(obj.Shapes),
            2,
            "TopoShapeList has wrong entry count (1): {0}".format(len(obj.Shapes)),
        )

        doc.undo()

        self.assertEqual(
            len(obj.Shapes),
            3,
            "TopoShapeList has wrong entry count (2): {0}".format(len(obj.Shapes)),
        )
        App.closeDocument(doc.Name)



#! python
# SPDX-License-Identifier: LGPL-2.1-or-later

# ShapeContent.py
# 2021, by Mark Ganson <TheMarkster>
# LGPL 2.1 or later

# this file is called from c++ in TaskCheckGeometry.cpp
# from buildShapeContent()

import FreeCAD as App
import Part

translate = App.Qt.translate


def roundVector(v, dec):
    return str([round(v[0], dec), round(v[1], dec), round(v[2], dec)])


def buildShapeContent(objArg, decimals=2, advancedShapeContent=True):
    linkName = ""
    if objArg.isDerivedFrom("App::Link"):
        linkName = "<" + objArg.Name + "> "

    obj = objArg
    shp = Part.getShape(objArg)
    typeStr = str(shp.ShapeType)
    lbl = "" if obj.Name == obj.Label else " (" + obj.Label + ")"
    result = linkName + obj.Name + lbl + "\n"
    result += translate("TaskCheckGeometryResults", "Shape type") + ":  " + typeStr + "\n"
    result += (
        translate("TaskCheckGeometryResults", "Vertices") + ":  " + str(len(shp.Vertexes)) + "\n"
    )
    result += translate("TaskCheckGeometryResults", "Edges") + ":  " + str(len(shp.Edges)) + "\n"
    result += translate("TaskCheckGeometryResults", "Wires") + ":  " + str(len(shp.Wires)) + "\n"
    result += translate("TaskCheckGeometryResults", "Faces") + ":  " + str(len(shp.Faces)) + "\n"
    result += translate("TaskCheckGeometryResults", "Shells") + ":  " + str(len(shp.Shells)) + "\n"
    result += translate("TaskCheckGeometryResults", "Solids") + ":  " + str(len(shp.Solids)) + "\n"
    result += (
        translate("TaskCheckGeometryResults", "CompSolids")
        + ":  "
        + str(len(shp.CompSolids))
        + "\n"
    )
    result += (
        translate("TaskCheckGeometryResults", "Compounds") + ":  " + str(len(shp.Compounds)) + "\n"
    )
    result += (
        translate("TaskCheckGeometryResults", "Shapes")
        + ":  "
        + str(
            len(
                shp.Vertexes
                + shp.Edges
                + shp.Wires
                + shp.Faces
                + shp.Shells
                + shp.Solids
                + shp.CompSolids
                + shp.Compounds
            )
        )
        + "\n"
    )
    if advancedShapeContent:
        result += "----------\n"
        if (
            hasattr(shp, "Area")
            and not "Wire" in typeStr
            and not "Edge" in typeStr
            and not "Vertex" in typeStr
        ):
            result += (
                translate("TaskCheckGeometryResults", "Area")
                + ":  "
                + str(round(shp.Area, decimals))
                + "\n"
            )
        if (
            hasattr(shp, "Volume")
            and not "Wire" in typeStr
            and not "Edge" in typeStr
            and not "Vertex" in typeStr
            and not "Face" in typeStr
        ):
            result += (
                translate("TaskCheckGeometryResults", "Volume")
                + ":  "
                + str(round(shp.Volume, decimals))
                + "\n"
            )
        if hasattr(shp, "Mass"):
            result += (
                translate("TaskCheckGeometryResults", "Mass")
                + ":  "
                + str(round(shp.Mass, decimals))
                + "\n"
            )
        if hasattr(shp, "Length"):
            result += (
                translate("TaskCheckGeometryResults", "Length")
                + ":  "
                + str(round(shp.Length, decimals))
                + "\n"
            )
        if hasattr(shp, "Curve") and hasattr(shp.Curve, "Radius"):
            result += (
                translate("TaskCheckGeometryResults", "Radius")
                + ":  "
                + str(round(shp.Curve.Radius, decimals))
                + "\n"
            )
        if hasattr(shp, "Curve") and hasattr(shp.Curve, "Center"):
            result += (
                translate("TaskCheckGeometryResults", "Curve center")
                + ":  "
                + str([round(vv, decimals) for vv in shp.Curve.Center])
                + "\n"
            )
        if hasattr(shp, "Curve") and hasattr(shp.Curve, "Continuity"):
            result += (
                translate("TaskCheckGeometryResults", "Continuity")
                + ":  "
                + str(shp.Curve.Continuity)
                + "\n"
            )
        if hasattr(shp, "CenterOfMass"):
            result += (
                translate("TaskCheckGeometryResults", "Center of mass")
                + ":  "
                + roundVector(shp.CenterOfMass, decimals)
                + "\n"
            )
        if hasattr(shp, "normalAt"):
            try:
                result += (
                    "normalAt(0):  " + str([round(vv, decimals) for vv in shp.normalAt(0)]) + "\n"
                )
            except Exception:
                try:
                    result += (
                        "normalAt(0,0):  "
                        + str([round(vv, decimals) for vv in shp.normalAt(0, 0)])
                        + "\n"
                    )
                except Exception:
                    pass
        if hasattr(shp, "isClosed") and ("Wire" in typeStr or "Edge" in typeStr):
            result += (
                translate("TaskCheckGeometryResults", "Is closed")
                + "?  "
                + str(shp.isClosed())
                + "\n"
            )
        if hasattr(shp, "Orientation"):
            result += (
                translate("TaskCheckGeometryResults", "Orientation")
                + ":  "
                + str(shp.Orientation)
                + "\n"
            )
        if hasattr(shp, "PrincipalProperties"):
            props = shp.PrincipalProperties
            for p in props:
                if isinstance(props[p], App.Vector) or isinstance(props[p], tuple):
                    result += str(p) + ":  " + roundVector(props[p], decimals) + "\n"
                else:
                    result += str(p) + ":  " + str(props[p]) + "\n"
        if hasattr(obj, "getGlobalPlacement"):
            if obj.getGlobalPlacement() != obj.Placement:
                rpl = obj.getGlobalPlacement() * obj.Placement.inverse()
                rot = rpl.Rotation
                if hasattr(shp, "CenterOfMass"):
                    result += (
                        translate("TaskCheckGeometryResults", "Global center of mass")
                        + ":  "
                        + roundVector(rpl.multVec(shp.CenterOfMass), decimals)
                        + "\n"
                    )
                if hasattr(shp, "PrincipalProperties"):
                    props = shp.PrincipalProperties
                    for p in props:
                        if "AxisOfInertia" in p:
                            result += (
                                "Global "
                                + str(p)
                                + ":  "
                                + roundVector(rot.multVec(props[p]), decimals)
                                + "\n"
                            )
            else:
                result += (
                    translate("TaskCheckGeometryResults", "Global placement")
                    + " = "
                    + translate("TaskCheckGeometryResults", "Placement")
                )
    return result



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
This file has the GUI command for checking and catching common errors in FreeCAD
CAM projects.
"""

from Path.Main.Sanity import Sanity
from PySide.QtCore import QT_TRANSLATE_NOOP
from PySide.QtGui import QFileDialog
import FreeCAD
import FreeCADGui
import Path
import Path.Log
import os
import webbrowser

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class CommandCAMSanity:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Sanity",
            "MenuText": QT_TRANSLATE_NOOP("CAM_Sanity", "Sanity Check"),
            "Accel": "P, S",
            "ToolTip": QT_TRANSLATE_NOOP("CAM_Sanity", "Checks the CAM job for common errors"),
        }

    def IsActive(self):
        selection = FreeCADGui.Selection.getSelectionEx()
        if len(selection) == 0:
            return False
        obj = selection[0].Object
        return isinstance(obj.Proxy, Path.Main.Job.ObjectJob)

    def Activated(self):
        FreeCADGui.addIconPath(":/icons")
        obj = FreeCADGui.Selection.getSelectionEx()[0].Object

        # Ask the user for a filename to save the report to

        defaultDir = os.path.split(FreeCAD.ActiveDocument.getFileName())[0]

        if defaultDir == "":
            defaultDir = os.path.expanduser("~")

        file_location = QFileDialog.getSaveFileName(
            None,
            translate("Path", "Save Sanity Check Report"),
            defaultDir,
            "HTML files (*.html)",
        )[0]

        if file_location == "":
            return

        sanity_checker = Sanity.CAMSanity(obj, file_location)
        html = sanity_checker.get_output_report()

        if html is None:
            Path.Log.error("Sanity check failed. No report generated.")
            return

        with open(file_location, "w") as fp:
            fp.write(html)

        FreeCAD.Console.PrintMessage("Sanity check report written to: {}\n".format(file_location))

        webbrowser.open_new_tab(file_location)


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_Sanity", CommandCAMSanity())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Frame command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Frame:
    "the Arch Frame command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Frame",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Frame", "Frame"),
            "Accel": "F, R",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Frame",
                "Creates a frame object from a planar 2D object (the extrusion path(s)) and a profile. Make sure objects are selected in that order.",
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        s = FreeCADGui.Selection.getSelection()
        if len(s) == 2:
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Frame"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.addModule("Draft")
            FreeCADGui.doCommand(
                "obj = Arch.makeFrame(FreeCAD.ActiveDocument."
                + s[0].Name
                + ",FreeCAD.ActiveDocument."
                + s[1].Name
                + ")"
            )
            FreeCADGui.doCommand("Draft.autogroup(obj)")
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()


FreeCADGui.addCommand("Arch_Frame", Arch_Frame())



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

__title__ = "BOPTools.ShapeMerge module"
__author__ = "DeepSOIC"
__url__ = "https://www.freecad.org"
__doc__ = "Tools for merging shapes with shared elements. Useful for final processing of results of Part.Shape.generalFuse()."

import Part
from .Utils import HashableShape


def findSharedElements(shape_list, element_extractor):
    if len(shape_list) < 2:
        raise ValueError(
            "findSharedElements: at least two shapes must be provided (have {num})".format(
                num=len(shape_list)
            )
        )

    all_elements = []  # list of sets of HashableShapes
    for shape in shape_list:
        all_elements.append(set([HashableShape(sh) for sh in element_extractor(shape)]))
    shared_elements = None
    for elements in all_elements:
        if shared_elements is None:
            shared_elements = elements
        else:
            shared_elements.intersection_update(elements)
    return [el.Shape for el in shared_elements]


def isConnected(shape1, shape2, shape_dim=-1):
    if shape_dim == -1:
        shape_dim = dimensionOfShapes([shape1, shape2])
    extractor = {
        0: None,
        1: (lambda sh: sh.Vertexes),
        2: (lambda sh: sh.Edges),
        3: (lambda sh: sh.Faces),
    }[shape_dim]
    return len(findSharedElements([shape1, shape2], extractor)) > 0


def splitIntoGroupsBySharing(list_of_shapes, element_extractor, split_connections=[]):
    """splitIntoGroupsBySharing(list_of_shapes, element_type, split_connections = []): find,
    which shapes in list_of_shapes are connected into groups by sharing elements.

    element_extractor: function that takes shape as input, and returns list of shapes.

    split_connections: list of shapes to exclude when testing for connections. Use to
    split groups on purpose.

    return: list of lists of shapes. Top-level list is list of groups; bottom level lists
    enumerate shapes of a group."""

    split_connections = set([HashableShape(element) for element in split_connections])

    groups = (
        []
    )  # list of tuples (shapes,elements). Shapes is a list of plain shapes. Elements is a set of HashableShapes - all elements of shapes in the group, excluding split_connections.

    # add shapes to the list of groups, one by one. If not connected to existing groups,
    # new group is created. If connected, shape is added to groups, and the groups are joined.
    for shape in list_of_shapes:
        shape_elements = set([HashableShape(element) for element in element_extractor(shape)])
        shape_elements.difference_update(split_connections)
        # search if shape is connected to any groups
        connected_to = []
        not_in_connected_to = []
        for iGroup in range(len(groups)):
            connected = False
            for element in shape_elements:
                if element in groups[iGroup][1]:
                    connected_to.append(iGroup)
                    connected = True
                    break
            else:
                # `break` not invoked, so `connected` is false
                not_in_connected_to.append(iGroup)

        # test if we need to join groups
        if len(connected_to) > 1:
            # shape bridges a gap between some groups. Join them into one.
            # rebuilding list of groups. First, add the new "supergroup", then add the rest
            groups_new = []

            supergroup = (list(), set())
            for iGroup in connected_to:
                supergroup[0].extend(groups[iGroup][0])  # merge lists of shapes
                supergroup[1].update(groups[iGroup][1])  # merge lists of elements
            groups_new.append(supergroup)

            l_groups = len(groups)
            groups_new.extend(
                [groups[i_group] for i_group in not_in_connected_to if i_group < l_groups]
            )
            groups = groups_new
            connected_to = [0]

        # add shape to the group it is connected to (if to many, the groups should have been unified by the above code snippet)
        if len(connected_to) > 0:
            iGroup = connected_to[0]
            groups[iGroup][0].append(shape)
            groups[iGroup][1].update(shape_elements)
        else:
            newgroup = ([shape], shape_elements)
            groups.append(newgroup)

    # done. Discard unnecessary data and return result.
    return [shapes for shapes, elements in groups]


def mergeSolids(
    list_of_solids_compsolids, flag_single=False, split_connections=[], bool_compsolid=False
):
    """mergeSolids(list_of_solids, flag_single = False): merges touching solids that share
    faces. If flag_single is True, it is assumed that all solids touch, and output is a
    single solid. If flag_single is False, the output is a compound containing all
    resulting solids.

    Note. CompSolids are treated as lists of solids - i.e., merged into solids."""

    solids = []
    for sh in list_of_solids_compsolids:
        solids.extend(sh.Solids)
    if flag_single:
        cs = Part.CompSolid(solids)
        return cs if bool_compsolid else Part.makeSolid(cs)
    else:
        if len(solids) == 0:
            return Part.Compound([])
        groups = splitIntoGroupsBySharing(solids, lambda sh: sh.Faces, split_connections)
        if bool_compsolid:
            merged_solids = [Part.CompSolid(group) for group in groups]
        else:
            merged_solids = [Part.makeSolid(Part.CompSolid(group)) for group in groups]
        return Part.makeCompound(merged_solids)


def mergeShells(list_of_faces_shells, flag_single=False, split_connections=[]):
    faces = []
    for sh in list_of_faces_shells:
        faces.extend(sh.Faces)
    if flag_single:
        return Part.makeShell(faces)
    else:
        groups = splitIntoGroupsBySharing(faces, lambda sh: sh.Edges, split_connections)
        return Part.makeCompound([Part.makeShell(group) for group in groups])


def mergeWires(list_of_edges_wires, flag_single=False, split_connections=[]):
    edges = []
    for sh in list_of_edges_wires:
        edges.extend(sh.Edges)
    if flag_single:
        return Part.Wire(edges)
    else:
        groups = splitIntoGroupsBySharing(edges, lambda sh: sh.Vertexes, split_connections)
        return Part.makeCompound([Part.Wire(Part.sortEdges(group)[0]) for group in groups])


def mergeVertices(list_of_vertices, flag_single=False, split_connections=[]):
    # no comprehensive support, just following the footprint of other mergeXXX()
    return Part.makeCompound(removeDuplicates(list_of_vertices))


def mergeShapes(list_of_shapes, flag_single=False, split_connections=[], bool_compsolid=False):
    """mergeShapes(list_of_shapes, flag_single = False, split_connections = [], bool_compsolid = False):
    merges list of edges/wires into wires, faces/shells into shells, solids/compsolids
    into solids or compsolids.

    list_of_shapes: shapes to merge. Shapes must share elements in order to be merged.

    flag_single: assume all shapes in list are connected. If False, return is a compound.
    If True, return is the single piece (e.g. a shell).

    split_connections: list of shapes that are excluded when searching for connections.
    This can be used for example to split a wire in two by supplying vertices where to
    split. If flag_single is True, this argument is ignored.

    bool_compsolid: determines behavior when dealing with solids/compsolids. If True,
    result is compsolid/compound of compsolids. If False, all touching solids and
    compsolids are unified into single solids. If not merging solids/compsolids, this
    argument is ignored."""

    if len(list_of_shapes) == 0:
        return Part.Compound([])
    args = [list_of_shapes, flag_single, split_connections]
    dim = dimensionOfShapes(list_of_shapes)
    if dim == 0:
        return mergeVertices(*args)
    elif dim == 1:
        return mergeWires(*args)
    elif dim == 2:
        return mergeShells(*args)
    elif dim == 3:
        args.append(bool_compsolid)
        return mergeSolids(*args)
    else:
        assert dim >= 0 and dim <= 3


def removeDuplicates(list_of_shapes):
    hashes = set()
    new_list = []
    for sh in list_of_shapes:
        hash = HashableShape(sh)
        if hash in hashes:
            pass
        else:
            new_list.append(sh)
            hashes.add(hash)
    return new_list


def dimensionOfShapes(list_of_shapes):
    """dimensionOfShapes(list_of_shapes): returns dimension (0D, 1D, 2D, or 3D) of shapes
    in the list. If dimension of shapes varies, TypeError is raised."""

    dimensions = [["Vertex"], ["Edge", "Wire"], ["Face", "Shell"], ["Solid", "CompSolid"]]
    dim = -1
    for sh in list_of_shapes:
        sht = sh.ShapeType
        for iDim in range(len(dimensions)):
            if sht in dimensions[iDim]:
                if dim == -1:
                    dim = iDim
                if iDim != dim:
                    raise TypeError(
                        "Shapes are of different dimensions ({t1} and {t2}), and cannot be merged or compared.".format(
                            t1=list_of_shapes[0].ShapeType, t2=sht
                        )
                    )
    return dim



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Label objects.

Labels are similar to text annotations but include a leader line
and an arrow in order to point to an object and indicate some of its
properties.
"""
## @package gui_labels
# \ingroup draftguitools
# \brief Provides GUI tools to create Label objects.

## \addtogroup draftguitools
# @{
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import DraftVecUtils
from draftguitools import gui_base_original
from draftguitools import gui_tool_utils
from draftguitools import gui_trackers as trackers
from draftutils import params
from draftutils.messages import _toolmsg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Label(gui_base_original.Creator):
    """Gui Command for the Label tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Label",
            "Accel": "D, L",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Label", "Label"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Label",
                "Creates a label, optionally attached to a selected object or subelement",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Label")
        self.ghost = None
        self.labeltype = params.get_param("labeltype")
        self.sel = Gui.Selection.getSelectionEx()
        if self.sel:
            self.sel = self.sel[0]
        self.ui.labelUi(title=translate("draft", self.featureName), callback=self.setmode)
        self.ui.xValue.setFocus()
        self.ui.xValue.selectAll()
        self.ghost = trackers.lineTracker()
        self.call = self.view.addEventCallback("SoEvent", self.action)
        _toolmsg(translate("draft", "Pick target point"))
        self.ui.isCopy.hide()

    def setmode(self, i):
        """Set the type of label, if it is associated to an object."""
        from draftobjects.label import get_label_types

        self.labeltype = get_label_types()[i]
        params.set_param("labeltype", self.labeltype)

    def finish(self, cont=False):
        """Finish the command."""
        self.end_callbacks(self.call)
        if self.ghost:
            self.ghost.finalize()
        super().finish()

    def create(self):
        """Create the actual object."""
        if len(self.node) == 3:
            targetpoint = self.node[0]
            basepoint = self.node[2]
            v = self.node[2].sub(self.node[1])
            dist = v.Length
            h = self.wp.u
            n = self.wp.axis
            r = self.wp.get_placement().Rotation

            if abs(DraftVecUtils.angle(v, h, n)) <= math.pi / 4:
                direction = "Horizontal"
                dist = -dist
            elif abs(DraftVecUtils.angle(v, h, n)) >= math.pi * 3 / 4:
                direction = "Horizontal"
            elif DraftVecUtils.angle(v, h, n) > 0:
                direction = "Vertical"
            else:
                direction = "Vertical"
                dist = -dist

            tp = DraftVecUtils.toString(targetpoint)
            sel = None
            if self.sel:
                sel = "FreeCAD.ActiveDocument." + self.sel.Object.Name

                if self.sel.SubElementNames:
                    sub = "'" + self.sel.SubElementNames[0] + "'"
                else:
                    sub = "None"

            pl = "FreeCAD.Placement"
            pl += "("
            pl += DraftVecUtils.toString(basepoint) + ", "
            pl += "FreeCAD.Rotation" + str(r.Q)
            pl += ")"

            Gui.addModule("Draft")
            _cmd = "Draft.make_label"
            _cmd += "("
            _cmd += "target_point=" + tp + ", "
            _cmd += "placement=" + pl + ", "
            if sel:
                _cmd += "target_object=" + sel + ", "
                _cmd += "subelements=" + sub + ", "
            _cmd += "label_type=" + "'" + self.labeltype + "'" + ", "
            # _cmd += "custom_text=" + "'Label'" + ", "
            _cmd += "direction=" + "'" + direction + "'" + ", "
            _cmd += "distance=" + str(dist)
            _cmd += ")"

            # Commit the creation instructions through the parent class,
            # the Creator class
            _cmd_list = [
                "_label_ = " + _cmd,
                "Draft.autogroup(_label_)",
                "FreeCAD.ActiveDocument.recompute()",
            ]
            self.commit(translate("draft", "Create Label"), _cmd_list)
        self.finish()

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif arg["Type"] == "SoLocation2Event":
            if hasattr(Gui, "Snapper"):
                Gui.Snapper.affinity = None  # don't keep affinity
            if len(self.node) == 2:
                gui_tool_utils.setMod(arg, gui_tool_utils.get_mod_constrain_key(), True)
            self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg)
            gui_tool_utils.redraw3DView()
        elif arg["Type"] == "SoMouseButtonEvent":
            if (arg["State"] == "DOWN") and (arg["Button"] == "BUTTON1"):
                if self.point:
                    self.ui.redraw()
                    if not self.node:
                        # first click
                        self.node.append(self.point)
                        self.ui.isRelative.show()
                        _toolmsg(translate("draft", "Pick endpoint of leader line"))
                        if self.planetrack:
                            self.planetrack.set(self.point)
                    elif len(self.node) == 1:
                        # second click
                        self.node.append(self.point)
                        if self.ghost:
                            self.ghost.p1(self.node[0])
                            self.ghost.p2(self.node[1])
                            self.ghost.on()
                        _toolmsg(translate("draft", "Pick text position"))
                    else:
                        # third click
                        self.node.append(self.point)
                        self.create()

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.point = App.Vector(numx, numy, numz)
        if not self.node:
            # first click
            self.node.append(self.point)
            self.ui.isRelative.show()
            _toolmsg(translate("draft", "Pick endpoint of leader line"))
            if self.planetrack:
                self.planetrack.set(self.point)
        elif len(self.node) == 1:
            # second click
            self.node.append(self.point)
            if self.ghost:
                self.ghost.p1(self.node[0])
                self.ghost.p2(self.node[1])
                self.ghost.on()
            _toolmsg(translate("draft", "Pick text position"))
        else:
            # third click
            self.node.append(self.point)
            self.create()


Draft_Label = Label
Gui.addCommand("Draft_Label", Label())

## @}



# ***************************************************************************
# *   Copyright (c) 2023 Peter McB                                          *
# *   added the function, mesh_2_femmesh, to convert the MESH               *
# *                     into a triangular FEMMESH                           *
# *                                                                         *
# *   Copyright (c) 2016 Frantisek Loeffelmann <LoffF@email.cz>             *
# *   extension to the work of:                                             *
#         Frantisek Loeffelmann, Ulrich Brammer, Bernd Hahnebach            *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FemMesh to Mesh converter"
__author__ = "Frantisek Loeffelmann, Ulrich Brammer, Bernd Hahnebach, Peter McB"
__url__ = "https://www.freecad.org"

## @package FwmMesh2Mesh
#  \ingroup FEM

import time

import FreeCAD
import Fem

# import Mesh


"""
from os.path import join
the_file = join(FreeCAD.getResourceDir(), "examples", "FEMExample.FCStd")
fc_file = FreeCAD.openDocument(the_file)
fem_mesh = fc_file.getObject("FEMMeshGmsh").FemMesh
result = fc_file.getObject("CCX_Results")
scale = 1  # displacement scale factor
from femmesh import femmesh2mesh
out_mesh = femmesh2mesh.femmesh_2_mesh(fem_mesh, result, scale)
import Mesh
Mesh.show(Mesh.Mesh(out_mesh))

"""
# These dictionaries list the nodes, that define faces of an element.
# The key is the face number, used internally by FreeCAD.
# The list contains the nodes in the element for each face.
tetFaces = {1: [0, 1, 2], 2: [0, 3, 1], 3: [1, 3, 2], 4: [2, 3, 0]}

pentaFaces = {1: [0, 1, 2], 2: [3, 5, 4], 3: [0, 3, 4, 1], 4: [1, 4, 5, 2], 5: [0, 2, 5, 3]}

hexaFaces = {  # hexa8 or hexa20 (ignoring mid-nodes)
    1: [0, 1, 2, 3],
    2: [4, 7, 6, 5],
    3: [0, 4, 5, 1],
    4: [1, 5, 6, 2],
    5: [2, 6, 7, 3],
    6: [3, 7, 4, 0],
}

pyraFaces = {  # pyra5 or pyra13 (ignoring mid-nodes)
    1: [0, 1, 2, 3],
    2: [0, 4, 1],
    3: [1, 4, 2],
    4: [2, 4, 3],
    5: [3, 4, 0],
}

face_dicts = {
    4: tetFaces,
    5: pyraFaces,
    6: pentaFaces,
    8: hexaFaces,
    10: tetFaces,
    13: pyraFaces,
    15: pentaFaces,
    20: hexaFaces,
}


def femmesh_2_mesh(myFemMesh, myResults=None, myDispScale=1):
    shiftBits = 20  # allows a million nodes, needs to be higher for more nodes in a FemMesh

    # This code generates a dict and a faceCode for each face of all elements
    # All faceCodes are than sorted.

    start_time = time.process_time()
    faceCodeList = []
    faceCodeDict = {}

    if myFemMesh.VolumeCount > 0:
        for ele in myFemMesh.Volumes:
            element_nodes = myFemMesh.getElementNodes(ele)
            # print("element_node: ", element_nodes)
            faceDef = face_dicts[len(element_nodes)]

            for key in faceDef:
                nodeList = []
                codeList = []
                faceCode = 0
                shifter = 0
                for nodeIdx in faceDef[key]:
                    nodeList.append(element_nodes[nodeIdx])
                    codeList.append(element_nodes[nodeIdx])
                codeList.sort()
                for node in codeList:
                    faceCode += node << shifter
                    # x << n: x shifted left by n bits = Multiplication
                    shifter += shiftBits
                # print("codeList: ", codeList)
                faceCodeDict[faceCode] = nodeList
                faceCodeList.append(faceCode)
    elif myFemMesh.FaceCount > 0:
        for ele in myFemMesh.Faces:
            element_nodes = myFemMesh.getElementNodes(ele)
            # print("element_node: ", element_nodes)
            if len(element_nodes) in [3, 6]:
                faceDef = {1: [0, 1, 2]}
            else:  # quad element
                faceDef = {1: [0, 1, 2, 3]}

            for key in faceDef:
                nodeList = []
                codeList = []
                faceCode = 0
                shifter = 0
                for nodeIdx in faceDef[key]:
                    nodeList.append(element_nodes[nodeIdx])
                    codeList.append(element_nodes[nodeIdx])
                codeList.sort()
                for node in codeList:
                    faceCode += node << shifter
                    # x << n: x shifted left by n bits = Multiplication
                    shifter += shiftBits
                # print("codeList: ", codeList)
                faceCodeDict[faceCode] = nodeList
                faceCodeList.append(faceCode)

    faceCodeList.sort()
    allFaces = len(faceCodeList)
    actFaceIdx = 0
    singleFaces = []
    # Here we search for faces, which do not have a counterpart.
    # These are the faces on the surface of the mesh.
    while actFaceIdx < allFaces:
        if actFaceIdx < (allFaces - 1):
            if faceCodeList[actFaceIdx] == faceCodeList[actFaceIdx + 1]:
                actFaceIdx += 2
            else:
                # print("found a single Face: ", faceCodeList[actFaceIdx])
                singleFaces.append(faceCodeList[actFaceIdx])
                actFaceIdx += 1
        else:
            FreeCAD.Console.PrintMessage(f"Found a last Face: {faceCodeList[actFaceIdx]}\n")
            singleFaces.append(faceCodeList[actFaceIdx])
            actFaceIdx += 1

    output_mesh = []
    if myResults:
        FreeCAD.Console.PrintMessage(f"{myResults.Name}\n")
        for myFace in singleFaces:
            face_nodes = faceCodeDict[myFace]
            dispVec0 = myResults.DisplacementVectors[myResults.NodeNumbers.index(face_nodes[0])]
            dispVec1 = myResults.DisplacementVectors[myResults.NodeNumbers.index(face_nodes[1])]
            dispVec2 = myResults.DisplacementVectors[myResults.NodeNumbers.index(face_nodes[2])]
            triangle = [
                myFemMesh.getNodeById(face_nodes[0]) + dispVec0 * myDispScale,
                myFemMesh.getNodeById(face_nodes[1]) + dispVec1 * myDispScale,
                myFemMesh.getNodeById(face_nodes[2]) + dispVec2 * myDispScale,
            ]
            output_mesh.extend(triangle)
            # print("my triangle: ", triangle)
            if len(face_nodes) == 4:
                dispVec3 = myResults.DisplacementVectors[myResults.NodeNumbers.index(face_nodes[3])]
                triangle = [
                    myFemMesh.getNodeById(face_nodes[2]) + dispVec2 * myDispScale,
                    myFemMesh.getNodeById(face_nodes[3]) + dispVec3 * myDispScale,
                    myFemMesh.getNodeById(face_nodes[0]) + dispVec0 * myDispScale,
                ]
                output_mesh.extend(triangle)
                # print("my 2. triangle: ", triangle)

    else:
        for myFace in singleFaces:
            face_nodes = faceCodeDict[myFace]
            triangle = [
                myFemMesh.getNodeById(face_nodes[0]),
                myFemMesh.getNodeById(face_nodes[1]),
                myFemMesh.getNodeById(face_nodes[2]),
            ]
            output_mesh.extend(triangle)
            # print("my triangle: ", triangle)
            if len(face_nodes) == 4:
                triangle = [
                    myFemMesh.getNodeById(face_nodes[2]),
                    myFemMesh.getNodeById(face_nodes[3]),
                    myFemMesh.getNodeById(face_nodes[0]),
                ]
                output_mesh.extend(triangle)
                # print("my 2. triangle: ", triangle)

    end_time = time.process_time()
    FreeCAD.Console.PrintMessage(f"Mesh by surface search method: {end_time - start_time}\n")
    # call to mesh_2_femmesh to convert mesh to femmesh before return statement
    mesh2femmesh = mesh_2_femmesh(myFemMesh, singleFaces, faceCodeDict)
    return output_mesh


# additional function to convert mesh to femmesh
def mesh_2_femmesh(myFemMesh, singleFaces, faceCodeDict):
    start_time = time.process_time()
    femmesh = Fem.FemMesh()
    myfemmesh = myFemMesh.Nodes
    # nodes contains the nodes that are used
    nodes = {}
    for myFace in singleFaces:
        face_nodes = faceCodeDict[myFace]
        for j in (0, 1, 2):
            try:
                nodes[face_nodes[j]] += 1
            except:
                nodes[face_nodes[j]] = 0
        if len(face_nodes) == 4:
            j = 3
            try:
                nodes[face_nodes[j]] += 1
            except:
                nodes[face_nodes[j]] = 0
    sfNode = femmesh.addNode
    sfFace = femmesh.addFace
    for key in myFemMesh.Nodes:
        mynode = myfemmesh[key]
        try:
            if nodes[key] >= 0:
                sfNode(mynode[0], mynode[1], mynode[2], key)
        except:
            pass

    output_mesh = []

    for myFace in singleFaces:
        face_nodes = faceCodeDict[myFace]
        sfFace(face_nodes[0], face_nodes[1], face_nodes[2])
        if len(face_nodes) == 4:
            sfFace(face_nodes[2], face_nodes[3], face_nodes[0])
    obj = FreeCAD.ActiveDocument.addObject("Fem::FemMeshObject", "Mesh2Fem")
    obj.FemMesh = femmesh
    end_time = time.process_time()
    FreeCAD.Console.PrintMessage(f"Convert to FemMesh: {end_time - start_time}\n")
    return obj


# end of mesh_2_femmesh



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Main.Stock as PathStock
from Path.Post.Processor import PostProcessor, PostProcessorFactory
import json

from FreeCAD import Units
from PySide import QtCore, QtGui


Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class JobPreferencesPage:
    def __init__(self, parent=None):
        import FreeCADGui

        self.form = FreeCADGui.PySideUic.loadUi(":preferences/PathJob.ui")
        self.form.toolBox.setCurrentIndex(0)  # Take that qt designer!

        self.postProcessorDefaultTooltip = self.form.defaultPostProcessor.toolTip()
        self.postProcessorArgsDefaultTooltip = self.form.defaultPostProcessorArgs.toolTip()
        self.processor = {}

    def saveSettings(self):
        jobTemplate = self.form.leDefaultJobTemplate.text()
        geometryTolerance = Units.Quantity(self.form.geometryTolerance.text())
        curveAccuracy = Units.Quantity(self.form.curveAccuracy.text())
        Path.Preferences.setJobDefaults(jobTemplate, geometryTolerance, curveAccuracy)

        if curveAccuracy:
            Path.Area.setDefaultParams(Accuracy=curveAccuracy)

        processor = str(self.form.defaultPostProcessor.currentText())
        args = str(self.form.defaultPostProcessorArgs.text())
        blacklist = []
        for i in range(0, self.form.postProcessorList.count()):
            item = self.form.postProcessorList.item(i)
            if item.checkState() == QtCore.Qt.CheckState.Unchecked:
                blacklist.append(item.text())
        Path.Preferences.setPostProcessorDefaults(processor, args, blacklist)

        path = str(self.form.leOutputFile.text())
        policy = str(self.form.cboOutputPolicy.currentText())
        Path.Preferences.setOutputFileDefaults(path, policy)
        self.saveStockSettings()

    def saveStockSettings(self):
        if self.form.stockGroup.isChecked():
            attrs = {}
            attrs["version"] = 1
            typ = [
                PathStock.StockType.CreateBox,
                PathStock.StockType.CreateCylinder,
                PathStock.StockType.FromBase,
            ][self.form.stock.currentIndex()]
            attrs["create"] = typ
            if typ == PathStock.StockType.CreateBox:
                attrs["length"] = FreeCAD.Units.Quantity(self.form.stockBoxLength.text()).UserString
                attrs["width"] = FreeCAD.Units.Quantity(self.form.stockBoxWidth.text()).UserString
                attrs["height"] = FreeCAD.Units.Quantity(self.form.stockBoxHeight.text()).UserString
            if typ == PathStock.StockType.CreateCylinder:
                attrs["radius"] = FreeCAD.Units.Quantity(
                    self.form.stockCylinderRadius.text()
                ).UserString
                attrs["height"] = FreeCAD.Units.Quantity(
                    self.form.stockCylinderHeight.text()
                ).UserString
            if typ == PathStock.StockType.FromBase:
                attrs["xneg"] = FreeCAD.Units.Quantity(self.form.stockExtXneg.text()).UserString
                attrs["xpos"] = FreeCAD.Units.Quantity(self.form.stockExtXpos.text()).UserString
                attrs["yneg"] = FreeCAD.Units.Quantity(self.form.stockExtYneg.text()).UserString
                attrs["ypos"] = FreeCAD.Units.Quantity(self.form.stockExtYpos.text()).UserString
                attrs["zneg"] = FreeCAD.Units.Quantity(self.form.stockExtZneg.text()).UserString
                attrs["zpos"] = FreeCAD.Units.Quantity(self.form.stockExtZpos.text()).UserString
            if self.form.stockPlacementGroup.isChecked():
                angle = FreeCAD.Units.Quantity(self.form.stockAngle.text()).Value
                axis = FreeCAD.Vector(
                    self.form.stockAxisX.value(),
                    self.form.stockAxisY.value(),
                    self.form.stockAxisZ.value(),
                )
                rot = FreeCAD.Rotation(axis, angle)
                attrs["rotX"] = rot.Q[0]
                attrs["rotY"] = rot.Q[1]
                attrs["rotZ"] = rot.Q[2]
                attrs["rotW"] = rot.Q[3]
                attrs["posX"] = FreeCAD.Units.Quantity(self.form.stockPositionX.text()).Value
                attrs["posY"] = FreeCAD.Units.Quantity(self.form.stockPositionY.text()).Value
                attrs["posZ"] = FreeCAD.Units.Quantity(self.form.stockPositionZ.text()).Value
            Path.Preferences.setDefaultStockTemplate(json.dumps(attrs))
        else:
            Path.Preferences.setDefaultStockTemplate("")

    def selectComboEntry(self, widget, text):
        index = widget.findText(text, QtCore.Qt.MatchFixedString)
        if index >= 0:
            widget.blockSignals(True)
            widget.setCurrentIndex(index)
            widget.blockSignals(False)

    def verifyAndUpdateDefaultPostProcessorWith(self, processor):
        defaultIsValid = False
        self.form.defaultPostProcessor.blockSignals(True)
        self.form.defaultPostProcessor.clear()
        self.form.defaultPostProcessor.addItem("")
        for i in range(0, self.form.postProcessorList.count()):
            item = self.form.postProcessorList.item(i)
            if item.checkState() == QtCore.Qt.CheckState.Checked:
                self.form.defaultPostProcessor.addItem(item.text())
                if item.text() == processor:
                    defaultIsValid = True
        # if we get here the default processor was disabled
        if not defaultIsValid:
            self.form.defaultPostProcessorArgs.setText("")
            processor = ""
        self.selectComboEntry(self.form.defaultPostProcessor, processor)
        self.form.defaultPostProcessor.blockSignals(False)

    def verifyAndUpdateDefaultPostProcessor(self):
        self.verifyAndUpdateDefaultPostProcessorWith(
            str(self.form.defaultPostProcessor.currentText())
        )

    def loadSettings(self):
        self.form.leDefaultJobTemplate.setText(Path.Preferences.defaultJobTemplate())

        blacklist = Path.Preferences.postProcessorBlacklist()
        for processor in Path.Preferences.allAvailablePostProcessors():
            item = QtGui.QListWidgetItem(processor)
            if processor in blacklist:
                item.setCheckState(QtCore.Qt.CheckState.Unchecked)
            else:
                item.setCheckState(QtCore.Qt.CheckState.Checked)
            item.setFlags(
                QtCore.Qt.ItemFlag.ItemIsSelectable
                | QtCore.Qt.ItemFlag.ItemIsEnabled
                | QtCore.Qt.ItemFlag.ItemIsUserCheckable
            )
            self.form.postProcessorList.addItem(item)
        self.verifyAndUpdateDefaultPostProcessorWith(Path.Preferences.defaultPostProcessor())

        self.form.defaultPostProcessorArgs.setText(Path.Preferences.defaultPostProcessorArgs())

        geomTol = Units.Quantity(Path.Preferences.defaultGeometryTolerance(), Units.Length)
        self.form.geometryTolerance.setText(geomTol.UserString)
        self.form.curveAccuracy.setText(
            Units.Quantity(Path.Preferences.defaultLibAreaCurveAccuracy(), Units.Length).UserString
        )

        self.form.leOutputFile.setText(Path.Preferences.defaultOutputFile())
        self.selectComboEntry(self.form.cboOutputPolicy, Path.Preferences.defaultOutputPolicy())

        self.form.tbDefaultJobTemplate.clicked.connect(self.browseDefaultJobTemplate)
        self.form.postProcessorList.itemEntered.connect(self.setProcessorListTooltip)
        self.form.postProcessorList.itemChanged.connect(self.verifyAndUpdateDefaultPostProcessor)
        self.form.defaultPostProcessor.currentIndexChanged.connect(
            self.updateDefaultPostProcessorToolTip
        )
        self.form.tbOutputFile.clicked.connect(self.browseOutputFile)

        self.loadStockSettings()

    def loadStockSettings(self):
        stock = Path.Preferences.defaultStockTemplate()
        index = -1
        if stock:
            attrs = json.loads(stock)
            if attrs.get("version") and 1 == int(attrs["version"]):
                stockType = attrs.get("create")
                if stockType == PathStock.StockType.FromBase:
                    index = 2
                elif stockType == PathStock.StockType.CreateBox:
                    index = 0
                elif stockType == PathStock.StockType.CreateCylinder:
                    index = 1
                else:
                    index = -1
        if -1 == index:
            attrs = {}
            self.form.stockGroup.setChecked(False)
        else:
            self.form.stockGroup.setChecked(True)
            self.form.stock.setCurrentIndex(index)

        # this either sets the default value or the value from the template for each field
        self.form.stockExtXneg.setText(attrs.get("xneg", "1 mm"))
        self.form.stockExtXpos.setText(attrs.get("xpos", "1 mm"))
        self.form.stockExtYneg.setText(attrs.get("yneg", "1 mm"))
        self.form.stockExtYpos.setText(attrs.get("ypos", "1 mm"))
        self.form.stockExtZneg.setText(attrs.get("zneg", "1 mm"))
        self.form.stockExtZpos.setText(attrs.get("zpos", "1 mm"))
        self.form.stockBoxLength.setText(attrs.get("length", "10 mm"))
        self.form.stockBoxWidth.setText(attrs.get("width", "10 mm"))
        self.form.stockBoxHeight.setText(attrs.get("height", "10 mm"))
        self.form.stockCylinderRadius.setText(attrs.get("radius", "5 mm"))
        self.form.stockCylinderHeight.setText(attrs.get("height", "10 mm"))

        posX = attrs.get("posX")
        posY = attrs.get("posY")
        posZ = attrs.get("posZ")
        rotX = attrs.get("rotX")
        rotY = attrs.get("rotY")
        rotZ = attrs.get("rotZ")
        rotW = attrs.get("rotW")
        if (
            posX is not None
            and posY is not None
            and posZ is not None
            and rotX is not None
            and rotY is not None
            and rotZ is not None
            and rotW is not None
        ):
            pos = FreeCAD.Vector(float(posX), float(posY), float(posZ))
            rot = FreeCAD.Rotation(float(rotX), float(rotY), float(rotZ), float(rotW))
            placement = FreeCAD.Placement(pos, rot)
            self.form.stockPlacementGroup.setChecked(True)
        else:
            placement = FreeCAD.Placement()
            self.form.stockPlacementGroup.setChecked(False)

        self.form.stockAngle.setText(
            FreeCAD.Units.Quantity("%f rad" % placement.Rotation.Angle).UserString
        )
        self.form.stockAxisX.setValue(placement.Rotation.Axis.x)
        self.form.stockAxisY.setValue(placement.Rotation.Axis.y)
        self.form.stockAxisZ.setValue(placement.Rotation.Axis.z)
        self.form.stockPositionX.setText(
            FreeCAD.Units.Quantity(placement.Base.x, FreeCAD.Units.Length).UserString
        )
        self.form.stockPositionY.setText(
            FreeCAD.Units.Quantity(placement.Base.y, FreeCAD.Units.Length).UserString
        )
        self.form.stockPositionZ.setText(
            FreeCAD.Units.Quantity(placement.Base.z, FreeCAD.Units.Length).UserString
        )

        self.setupStock(index)
        self.form.stock.currentIndexChanged.connect(self.setupStock)

    def setupStock(self, index):
        if 0 == index:
            self.form.stockFromBase.hide()
            self.form.stockCreateBox.show()
            self.form.stockCreateCylinder.hide()
        elif 1 == index:
            self.form.stockFromBase.hide()
            self.form.stockCreateBox.hide()
            self.form.stockCreateCylinder.show()
        else:
            self.form.stockFromBase.show()
            self.form.stockCreateBox.hide()
            self.form.stockCreateCylinder.hide()

    def getPostProcessor(self, name):
        if not name in self.processor:
            processor = PostProcessorFactory.get_post_processor(None, name)
            self.processor[name] = processor
            return processor
        return self.processor[name]

    def setPostProcessorTooltip(self, widget, name, default):
        processor = self.getPostProcessor(name)
        if processor.tooltip:
            widget.setToolTip(processor.tooltip)
        else:
            widget.setToolTip(default)

    def setProcessorListTooltip(self, item):
        self.setPostProcessorTooltip(self.form.postProcessorList, item.text(), "")

    def updateDefaultPostProcessorToolTip(self):
        name = str(self.form.defaultPostProcessor.currentText())
        if name:
            self.setPostProcessorTooltip(
                self.form.defaultPostProcessor, name, self.postProcessorDefaultTooltip
            )
            processor = self.getPostProcessor(name)
            if processor.tooltipArgs:
                self.form.defaultPostProcessorArgs.setToolTip(processor.tooltipArgs)
            else:
                self.form.defaultPostProcessorArgs.setToolTip(self.postProcessorArgsDefaultTooltip)
        else:
            self.form.defaultPostProcessor.setToolTip(self.postProcessorDefaultTooltip)
            self.form.defaultPostProcessorArgs.setToolTip(self.postProcessorArgsDefaultTooltip)

    def bestGuessForFilePath(self):

        path = Path.Preferences.defaultFilePath()
        if not path:
            path = Path.Preferences.filePath()
        return path

    def browseDefaultJobTemplate(self):
        path = self.form.leDefaultJobTemplate.text()
        if not path:
            path = self.bestGuessForFilePath()
        foo = QtGui.QFileDialog.getOpenFileName(
            QtGui.QApplication.activeWindow(), "Path - Job Template", path, "job_*.json"
        )[0]
        if foo:
            self.form.leDefaultJobTemplate.setText(foo)

    def browseOutputFile(self):
        path = self.form.leOutputFile.text()
        foo = QtGui.QFileDialog.getExistingDirectory(
            QtGui.QApplication.activeWindow(), "Path - Output File/Directory", path
        )
        if foo:
            self.form.leOutputFile.setText(foo)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Layer object."""
## @package layer
# \ingroup draftobjects
# \brief Provides the object code for the Layer object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftutils import gui_utils
from draftutils import utils
from draftutils.messages import _log


class Layer:
    """The Layer object.

    This class is normally used to extend a base `App::FeaturePython` object.
    """

    def __init__(self, obj):
        self.Type = "Layer"
        self.Object = obj
        self.set_properties(obj)

        obj.Proxy = self

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        if "Group" not in obj.PropertiesList:
            _tip = QT_TRANSLATE_NOOP("App::Property", "The objects that are part of this layer")
            # "App::PropertyLinkListHidden" instead of "App::PropertyLinkList" has 2 advantages:
            # 1. No 'might break' warning when deleting an object nested in a layer.
            # 2. No 'out of scope' warning for objects also nested in f.e. a Std_Part.
            obj.addProperty("App::PropertyLinkListHidden", "Group", "Layer", _tip, locked=True)

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored."""
        # Group property type changed in v1.0:
        if obj.getTypeIdOfProperty("Group") != "App::PropertyLinkListHidden":
            grp = obj.Group  # Type: "App::PropertyLinkList".
            group_removed = obj.removeProperty("Group")  # Not possible for VisGroups (< v0.19)
            self.set_properties(obj)
            if group_removed:
                obj.Group = grp
                _log("v1.0, " + obj.Name + ", changed 'Group' property type")

        gui_utils.restore_view_object(
            obj, vp_module="view_layer", vp_class="ViewProviderLayer", format=False
        )

        if not getattr(obj, "ViewObject", None):
            return
        vobj = obj.ViewObject

        if self.Type == "VisGroup":  # Type prior to v0.19.
            self.Type = "Layer"
            # It is not possible to change the property group of obj.Group.
            for prop in ("DrawStyle", "LineColor", "LineWidth", "ShapeColor", "Transparency"):
                vobj.setGroupOfProperty(prop, "Layer")

        if not hasattr(vobj, "ShapeAppearance"):
            vobj.Proxy.set_properties(vobj)
            material = App.Material()  #  Material with default v0.21 properties.
            material.DiffuseColor = vobj.ShapeColor
            material.Transparency = vobj.Transparency / 100
            vobj.ShapeAppearance = (material,)
            vobj.setPropertyStatus("ShapeColor", "Hidden")
            if hasattr(vobj, "OverrideShapeColorChildren"):  # v0.19 - v0.21
                vobj.OverrideShapeAppearanceChildren = vobj.OverrideShapeColorChildren
                vobj.removeProperty("OverrideShapeColorChildren")
            _log("v1.0, " + obj.Name + ", updated view properties")

    def dumps(self):
        """Return a tuple of objects to save or None."""
        return self.Type

    def loads(self, state):
        """Set the internal properties from the restored state."""
        if state:
            self.Type = state

    def execute(self, obj):
        """Execute when the object is created or recomputed. Do nothing."""
        pass

    def _get_other_layers(self, obj, child):
        other_lyrs = []
        for find in child.Document.findObjects(Type="App::FeaturePython"):
            if utils.get_type(find) == "Layer" and find != obj and child in find.Group:
                other_lyrs.append(find)
        return other_lyrs

    def onBeforeChange(self, obj, prop):
        if prop == "Group":
            self.oldGroup = obj.Group

    def onChanged(self, obj, prop):
        if prop != "Group":
            return
        vobj = getattr(obj, "ViewObject", None)
        old_grp = getattr(self, "oldGroup", [])
        for child in obj.Group:
            if child in old_grp:
                continue
            for other_lyr in self._get_other_layers(obj, child):
                other_grp = other_lyr.Group
                other_grp.remove(child)
                other_lyr.Group = other_grp
            if vobj is None:
                continue
            for prop in ("LineColor", "ShapeAppearance", "LineWidth", "DrawStyle", "Visibility"):
                vobj.Proxy.change_view_properties(vobj, prop, old_prop=None, targets=[child])

    def addObject(self, obj, child):
        if utils.get_type(child) in ("Layer", "LayerContainer"):
            return
        grp = obj.Group
        if child in grp:
            return
        grp.append(child)
        obj.Group = grp

    def removeObject(self, obj, child):
        grp = obj.Group
        if not child in grp:
            return
        grp.remove(child)
        obj.Group = grp


# Alias for compatibility with v0.18 and earlier
_VisGroup = Layer


class LayerContainer:
    """The container object for layers.

    This class is normally used to extend
    a base `App::DocumentObjectGroupPython` object.
    """

    def __init__(self, obj):
        self.Type = "LayerContainer"
        obj.Proxy = self

    def onDocumentRestored(self, obj):
        gui_utils.restore_view_object(
            obj, vp_module="view_layer", vp_class="ViewProviderLayerContainer", format=False
        )

    def execute(self, obj):
        """Execute when the object is created or recomputed.

        Update the value of `Group` by sorting the contained layers
        by `Label`.
        """
        grp = obj.Group
        grp.sort(key=lambda layer: layer.Label)
        obj.Group = grp

    def dumps(self):
        """Return a tuple of objects to save or None."""
        if hasattr(self, "Type"):
            return self.Type

    def loads(self, state):
        """Set the internal properties from the restored state."""
        if state:
            self.Type = state


def get_layer(obj):
    """Get the layer the object belongs to."""
    finds = obj.Document.findObjects(Name="LayerContainer")
    if not finds:
        return None
    # First look in the LayerContainer:
    for layer in finds[0].Group:
        if utils.get_type(layer) == "Layer" and obj in layer.Group:
            return layer
    # If not found, look through all App::FeaturePython objects (not just layers):
    for find in obj.Document.findObjects(Type="App::FeaturePython"):
        if utils.get_type(find) == "Layer" and obj in find.Group:
            return find
    return None


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Define geometry functions for manipulating shapes in the Draft Workbench.

These functions are used by different object creation functions
of the Draft Workbench, both in `Draft.py` and `DraftTools.py`.
They operate on the internal shapes (`Part::TopoShape`) of different objects
and on their subelements, that is, vertices, edges, and faces.
"""
## \defgroup DRAFTGEOMUTILS DraftGeomUtils
#  \ingroup UTILITIES
#  \brief Shape manipulation utilities for the Draft workbench
#
# Shapes manipulation utilities

## \addtogroup DRAFTGEOMUTILS
#  @{

__title__ = "FreeCAD Draft Workbench - Geometry library"
__author__ = "Yorik van Havre, Jacques-Antoine Gaudin, Ken Cline"
__url__ = ["https://www.freecad.org"]

# Doesn't need requisites
from draftgeoutils.linear_algebra import linearFromPoints, determinant

# Needs math, Part, and vector tools
from draftgeoutils.general import NORM

from draftgeoutils.general import (
    precision,
    vec,
    edg,
    getVerts,
    v1,
    isNull,
    isPtOnEdge,
    hasCurves,
    isAligned,
    getQuad,
    areColinear,
    hasOnlyWires,
    geomType,
    isValidPath,
    findClosest,
    getBoundaryAngles,
)

# Needs general functions
from draftgeoutils.geometry import (
    findPerpendicular,
    findDistance,
    getSplineNormal,
    get_spline_normal,
    getNormal,
    get_normal,
    get_shape_normal,
    getRotation,
    isPlanar,
    is_planar,
    calculatePlacement,
    mirror,
    are_coplanar,
    is_straight_line,
    mirror_matrix,
    uv_vectors_from_face,
    placement_from_face,
    placement_from_points,
    distance_to_plane,
    project_point_on_plane,
)

from draftgeoutils.edges import (
    findEdge,
    orientEdge,
    isSameLine,
    isLine,
    is_line,
    invert,
    findMidpoint,
    getTangent,
    get_referenced_edges,
)

from draftgeoutils.faces import (
    concatenate,
    getBoundary,
    isCoplanar,
    is_coplanar,
    bind,
    cleanFaces,
    removeSplitter,
)

from draftgeoutils.arcs import isClockwise, isWideAngle, arcFrom2Pts, arcFromSpline

from draftgeoutils.cuboids import isCubic, getCubicDimensions

# Needs geometry functions
from draftgeoutils.circle_inversion import pointInversion, polarInversion, circleInversion

# Needs edges functions
from draftgeoutils.sort_edges import sortEdges, sortEdgesOld

from draftgeoutils.intersections import findIntersection, wiresIntersect, connect, angleBisection

from draftgeoutils.wires import (
    findWires,
    findWiresOld,
    findWiresOld2,
    flattenWire,
    superWire,
    isReallyClosed,
    curvetowire,
    curvetosegment,
    rebaseWire,
    removeInterVertices,
    cleanProjection,
    tessellateProjection,
    get_placement_perpendicular_to_wire,
    get_extended_wire,
)

# Needs wires functions
from draftgeoutils.fillets import fillet, filletWire

# Needs intersections functions
from draftgeoutils.offsets import pocket2d, offset, offsetWire

from draftgeoutils.circles import (
    findClosestCircle,
    getCircleFromSpline,
    circlefrom1Line2Points,
    circlefrom2Lines1Point,
    circleFrom2LinesRadius,
    circleFrom3LineTangents,
    circleFromPointLineRadius,
    circleFrom2PointsRadius,
    findHomotheticCenterOfCircles,
    findRadicalAxis,
    findRadicalCenter,
)

from draftgeoutils.circles_apollonius import (
    outerSoddyCircle,
    innerSoddyCircle,
    circleFrom3CircleTangents,
)

# Needs circles_apollonius functions
# These functions are not imported because they are incomplete;
# they require pre-requisite functions that haven't been written
# from draftgeoutils.circles_incomplete import (circleFrom2tan1pt,
#                                               circleFrom2tan1rad,
#                                               circleFrom1tan2pt,
#                                               circleFrom1tan1pt1rad,
#                                               circleFrom3tan)

##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2020 Carlo Pavan <carlopav@gmail.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Modules that contain functions for use with scripted objects and commands.

These functions provide support for dealing with the custom objects
defined within the workbench.
The functions are meant to be used in the creation step of the objects,
by the functions in the `draftmake` package, but also by the graphical
GuiCommands in the `draftguitools` and `drafttaskpanels` packages.

These functions should deal with the internal shapes of the objects,
and their special properties. They should not be very generic;
if they are very generic then they are more appropriate to be included
in the `draftutils` package.

These functions, together with those defined in the `draftmake` package,
represent the public application programming interface (API)
of the Draft Workbench, and should be made available in the `Draft`
namespace by importing them in the `Draft` module.
"""
## \defgroup draftfunctions draftfunctions
# \ingroup DRAFT
# \brief Modules with functions for use with scripted objects and GuiCommands.



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Vector

import Part
from BOPTools import SplitFeatures

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Constraint Tie",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["fixed", "force", "tie"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.constraint_tie import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=42783

constraint tie, bond two surfaces together (solid mesh only)

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects
    # cones cut
    cone_outer_sh = Part.makeCone(1100, 1235, 1005, Vector(0, 0, 0), Vector(0, 0, 1), 359)
    cone_inner_sh = Part.makeCone(1050, 1185, 1005, Vector(0, 0, 0), Vector(0, 0, 1), 359)
    cone_cut_sh = cone_outer_sh.cut(cone_inner_sh)
    cone_cut_obj = doc.addObject("Part::Feature", "Cone_Cut")
    cone_cut_obj.Shape = cone_cut_sh

    # lines
    line_fix_sh = Part.Edge(Part.LineSegment(Vector(0, -1235, 1005), Vector(0, -1185, 1005)))
    line_fix_obj = doc.addObject("Part::Feature", "Line_Fix")
    line_fix_obj.Shape = line_fix_sh
    line_force_sh = Part.Edge(Part.LineSegment(Vector(0, 1185, 1005), Vector(0, 1235, 1005)))
    line_force_obj = doc.addObject("Part::Feature", "Line_Force")
    line_force_obj.Shape = line_force_sh

    geom_obj = SplitFeatures.makeBooleanFragments(name="BooleanFragments")
    geom_obj.Objects = [cone_cut_obj, line_fix_obj, line_force_obj]
    if FreeCAD.GuiUp:
        cone_cut_obj.ViewObject.hide()
        line_fix_obj.ViewObject.hide()
        line_force_obj.ViewObject.hide()
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
        solver_obj.SplitInputWriter = False
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "MechanicalMaterial")
    mat = material_obj.Material
    mat["Name"] = "Calculix-Steel"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, "Edge1")]
    analysis.addObject(con_fixed)

    # constraint force
    con_force = ObjectsFem.makeConstraintForce(doc, "ConstraintForce")
    con_force.References = [(geom_obj, "Edge2")]
    con_force.Force = "10000.0 N"  # 10 kN
    con_force.Direction = (geom_obj, ["Edge2"])
    con_force.Reversed = False
    analysis.addObject(con_force)

    # constraint tie
    con_tie = ObjectsFem.makeConstraintTie(doc, "ConstraintTie")
    con_tie.References = [
        (geom_obj, "Face5"),
        (geom_obj, "Face7"),
    ]
    con_tie.Tolerance = 25.0
    analysis.addObject(con_tie)

    # mesh
    from .meshes.mesh_constraint_tie_tetra10 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False

    doc.recompute()
    return doc



# ***************************************************************************
# *   Copyright (c) 2020 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FEM Utilities"
__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"

tokens = (
    "NAME",
    "FLOAT",
    "INT",
    "PLUS",
    "MINUS",
    "TIMES",
    "DIVIDE",
    "EQUALS",
    "LPAREN",
    "RPAREN",
    "POWER",
)

# Tokens

t_PLUS = r"\+"
t_MINUS = r"-"
t_TIMES = r"\*"
t_DIVIDE = r"/"
t_EQUALS = r"="
t_LPAREN = r"\("
t_RPAREN = r"\)"
t_POWER = r"\^"
t_NAME = r"[a-zA-Z_][a-zA-Z0-9_]*"


def t_FLOAT(t):
    r"\d+\.(\d+)?([eE][-+]?\d+)?"
    t.value = float(t.value)
    return t


def t_INT(t):
    r"\d+"
    t.value = int(t.value)
    return t


# Ignored characters
t_ignore = " \t"


def t_COMMENT(t):
    r"\#.*"
    pass


def t_newline(t):
    r"\n+"
    t.lexer.lineno += t.value.count("\n")


def t_error(t):
    print("Illegal character '%s'" % t.value[0])
    t.lexer.skip(1)


# Build the lexer
import ply.lex as lex

lex.lex()


# Precedence rules for the arithmetic operators
precedence = (
    ("left", "PLUS", "MINUS"),
    ("left", "TIMES", "DIVIDE"),
    ("left", "POWER"),
    ("right", "UMINUS"),
)

# dictionary of names (for storing variables)
names = {}


def p_statement_assign(p):
    "statement : NAME EQUALS expression"
    names[p[1]] = p[3]


def p_statement_expr(p):
    "statement : expression"
    print(p[1])


def p_expression_binop(p):
    """expression : expression PLUS expression
    | expression MINUS expression
    | expression TIMES expression
    | expression DIVIDE expression
    | expression POWER expression"""
    if p[2] == "+":
        p[0] = p[1] + p[3]
    elif p[2] == "-":
        p[0] = p[1] - p[3]
    elif p[2] == "*":
        p[0] = p[1] * p[3]
    elif p[2] == "/":
        p[0] = p[1] / p[3]
    elif p[2] == "^":
        p[0] = p[1] ** p[3]


def p_expression_uminus(p):
    "expression : MINUS expression %prec UMINUS"
    p[0] = -p[2]


def p_expression_group(p):
    "expression : LPAREN expression RPAREN"
    p[0] = p[2]


def p_expression_float(p):
    "expression : FLOAT"
    p[0] = p[1]


def p_expression_int(p):
    "expression : INT"
    p[0] = p[1]


def p_expression_name(p):
    "expression : NAME"
    try:
        p[0] = names[p[1]]
    except LookupError:
        print("Undefined name '%s'" % p[1])
        p[0] = 0


def p_error(p):
    print("Syntax error at '%s'" % p.value)


import ply.yacc as yacc

yacc.yacc(debug=False, write_tables=False)
